/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Render
Version:   1.0
ID:        ID_RENDER
Status:    Stable
Category:  GUI
Author:    Rocklyte Systems
Date:      June 2004
Copyright: Rocklyte Systems (c) 2000-2004.  All rights reserved.
Keywords:  graphics, render, area, graphic, box, rectangle, window, layer
Short:     Manages the display and positioning of 2-Dimensional rendered graphics.

###DESCRIPTION###
<p>The Render class is the primary tool that is used for interface
construction.  On its own the Render class does very little in terms of
creating graphics, but it has the sole responsibility of managing 2-Dimensional
graphics space and keeping everything organised for the display.</p>

<p>Graphically, each render object represents a rectangular area of space
on the display.  The biggest render object, often referred to as the 'Master'
represents the screen space, and is as big as the display itself, if not larger.
render objects can be placed inside the master and are typically known as
children.  You can place more render objects inside of these children,
causing a hierachy to develop that consists of many objects that are all
working together on the display.</p>

<p>In order to actually draw graphics onto the display, classes that
have been specifically created for drawing graphics must be used to create
a meaningful interface.  Classes such as Box, Image, Text and Gradient are
able to create an assemblance of imagery that has meaning to the user.
The interface is then enhanced through the use of user-interactive classes
such as OnClick, Movement and Pointer.  With a little effort and imagination,
a customised interface can thus be assembled and presented to the user
without much difficulty on the part of the developer.</p>

<p>While this is a simple concept, the Render class is at the base of its foundation and
is one of the largest and most important classes in the system.  It provides a
great deal of functionality which cannot be summarised in this introduction,
but you will find a lot of technical detail on each individual field in this
manual.  The DML Whitepaper has a good summary as to the basic use of the
Render class.  It is also recommended that you refer to the DML scripts that
come with Athene if you require a real-world view of how the class is
typically used.</p>

###END###

Backing Stores
--------------
The Render class uses the "backing store" technique for preserving the graphics
of rendered areas.  This is an absolute requirement when drawing regions at
points where processes intersect in the layer hierarchy.  Not only does it
speed up exposes, but it is the only reasonable way to get masking and
translucency effects working correctly.  Note that backing store graphics are
stored in public Bitmaps so that processes can share graphics information
without having to communicate with each other directly.

*****************************************************************************/

static ERROR RENDER_ActionNotify(struct Render *, struct acActionNotify *);
static ERROR RENDER_Activate(struct Render *, APTR);
static ERROR RENDER_ClosingTag(struct Render *, APTR);
static ERROR RENDER_Disable(struct Render *, APTR);
static ERROR RENDER_Draw(struct Render *, struct acDraw *);
static ERROR RENDER_Enable(struct Render *, APTR);
static ERROR RENDER_EventMessage(struct Render *, struct acEventMessage *);
static ERROR RENDER_Focus(struct Render *, APTR);
static ERROR RENDER_Free(struct Render *, APTR);
static ERROR RENDER_Hide(struct Render *, APTR);
static ERROR RENDER_Init(struct Render *, APTR);
static ERROR RENDER_LostFocus(struct Render *, APTR);
static ERROR RENDER_Move(struct Render *, struct acMove *);
static ERROR RENDER_MoveToBack(struct Render *, APTR);
static ERROR RENDER_MoveToFront(struct Render *, APTR);
static ERROR RENDER_MoveToPoint(struct Render *, struct acMoveToPoint *);
static ERROR RENDER_NewOwner(struct Render *, struct acNewOwner *);
static ERROR RENDER_NewObject(struct Render *, APTR);
static ERROR RENDER_Redimension(struct Render *, struct acRedimension *);
static ERROR RENDER_Resize(struct Render *, struct acResize *);
static ERROR RENDER_SaveImage(struct Render *, struct acSaveImage *);
static ERROR RENDER_Scroll(struct Render *, struct acScroll *);
static ERROR RENDER_ScrollToPoint(struct Render *, struct acScrollToPoint *);
static ERROR RENDER_Show(struct Render *, APTR);
static ERROR RENDER_UserClick(struct Render *, struct acUserClick *);
static ERROR RENDER_UserClickRelease(struct Render *, struct acUserClickRelease *);
static ERROR RENDER_UserMovement(struct Render *, struct acUserMovement *);

/*** Get Functions ***/

static ERROR GET_AbsX(struct Render *, LONG *);
static ERROR GET_AbsY(struct Render *, LONG *);
static ERROR GET_AmtColours(struct Render *, LONG *);
static ERROR GET_Bitmap(struct Render *, struct Bitmap **);
static ERROR GET_BitsPerPixel(struct Render *, LONG *);
static ERROR GET_Bottom(struct Render *, LONG *);
static ERROR GET_Buffer(struct Render *, LONG *);
static ERROR GET_BytesPerPixel(struct Render *, LONG *);
static ERROR GET_ColourRGB(struct Render *, struct RGB *);
static ERROR GET_FastResize(struct Render *, LONG *);
static ERROR GET_Height(struct Render *, struct Variable *);
static ERROR GET_InsideHeight(struct Render *, LONG *);
static ERROR GET_InsideWidth(struct Render *, LONG *);
static ERROR GET_Opacity(struct Render *, FLOAT *);
static ERROR GET_Precopy(struct Render *, LONG *);
static ERROR GET_Region(struct Render *, LONG *);
static ERROR GET_Right(struct Render *, LONG *);
static ERROR GET_Host(struct Render *, LONG *);
static ERROR GET_Sticky(struct Render *, LONG *);
static ERROR GET_Invisible(struct Render *, LONG *);
static ERROR GET_UserFocus(struct Render *, OBJECTID *);
static ERROR GET_Visible(struct Render *, LONG *);
static ERROR GET_VisibleHeight(struct Render *, LONG *);
static ERROR GET_VisibleWidth(struct Render *, LONG *);
static ERROR GET_Width(struct Render *, struct Variable *);
static ERROR GET_WindowHandle(struct Render *, WINHANDLE *);
static ERROR GET_XCoord(struct Render *, struct Variable *);
static ERROR GET_XOffset(struct Render *, struct Variable *);
static ERROR GET_YCoord(struct Render *, struct Variable *);
static ERROR GET_YOffset(struct Render *, struct Variable *);

/*** Set Functions ***/

static ERROR SET_AbsX(struct Render *, LONG *);
static ERROR SET_AbsY(struct Render *, LONG *);
static ERROR SET_BitsPerPixel(struct Render *, LONG *);
static ERROR SET_BottomMargin(struct Render *, LONG *);
static ERROR SET_Buffer(struct Render *, LONG *);
static ERROR SET_Host(struct Render *, LONG *);
static ERROR SET_Colour(struct Render *, STRING);
static ERROR SET_ColourRGB(struct Render *, struct RGB *);
static ERROR SET_FastResize(struct Render *, LONG *);
static ERROR SET_Flags(struct Render *, LONG *);
static ERROR SET_Frame(struct Render *, LONG *);
static ERROR SET_Hide(struct Render *, LONG *);
static ERROR SET_InsideHeight(struct Render *, LONG *);
static ERROR SET_InsideWidth(struct Render *, LONG *);
static ERROR SET_LeftLimit(struct Render *, LONG *);
static ERROR SET_TopLimit(struct Render *, LONG *);
static ERROR SET_BottomLimit(struct Render *, LONG *);
static ERROR SET_RightLimit(struct Render *, LONG *);
static ERROR SET_MaxHeight(struct Render *, LONG *);
static ERROR SET_MaxWidth(struct Render *, LONG *);
static ERROR SET_MinHeight(struct Render *, LONG *);
static ERROR SET_MinWidth(struct Render *, LONG *);
static ERROR SET_Movement(struct Render *, LONG *);
static ERROR SET_Opacity(struct Render *, FLOAT *);
static ERROR SET_Precopy(struct Render *, LONG *);
static ERROR SET_Region(struct Render *, LONG *);
static ERROR SET_RightMargin(struct Render *, LONG *);
static ERROR SET_RouteFocus(struct Render *, OBJECTID *);
static ERROR SET_StickToBack(struct Render *, LONG *);
static ERROR SET_StickToFront(struct Render *, LONG *);
static ERROR SET_Sticky(struct Render *, LONG *);
static ERROR SET_Invisible(struct Render *, LONG *);
static ERROR SET_Visible(struct Render *, LONG *);
static ERROR SET_WindowHandle(struct Render *, WINHANDLE *);

static ERROR SET_Height(struct Render *, struct Variable *);
static ERROR SET_Width(struct Render *, struct Variable *);
static ERROR SET_XCoord(struct Render *, struct Variable *);
static ERROR SET_XOffset(struct Render *, struct Variable *);
static ERROR SET_YOffset(struct Render *, struct Variable *);
static ERROR SET_YCoord(struct Render *, struct Variable *);

/*** Movement Flags ***/

#define MOVE_VERTICAL   0x0001
#define MOVE_HORIZONTAL 0x0002

static struct FieldDef MovementFlags[] = {
   { "VERTICAL",   MOVE_VERTICAL },
   { "HORIZONTAL", MOVE_HORIZONTAL },
   { NULL, NULL }
};

static struct FieldDef AlignFlags[] = {
   { "RIGHT",    ALIGN_RIGHT    }, { "LEFT",       ALIGN_LEFT    },
   { "BOTTOM",   ALIGN_BOTTOM   }, { "TOP",        ALIGN_TOP     },
   { "CENTER",   ALIGN_CENTER   }, { "MIDDLE",     ALIGN_MIDDLE  },
   { "VERTICAL", ALIGN_VERTICAL }, { "HORIZONTAL", ALIGN_HORIZONTAL },
   { NULL, NULL }
};

static struct FieldDef CursorLookup[] = { /* This table is copied from pointer_class.c */
   { "None",            0 },
   { "Default",         PTR_DEFAULT },
   { "SizeBottomLeft",  PTR_SIZE_BOTTOM_LEFT },
   { "SizeBottomRight", PTR_SIZE_BOTTOM_RIGHT },
   { "SizeTopLeft",     PTR_SIZE_TOP_LEFT },
   { "SizeTopRight",    PTR_SIZE_TOP_RIGHT },
   { "SizeLeft",        PTR_SIZE_LEFT },
   { "SizeRight",       PTR_SIZE_RIGHT },
   { "SizeTop",         PTR_SIZE_TOP },
   { "SizeBottom",      PTR_SIZE_BOTTOM },
   { "Crosshair",       PTR_CROSSHAIR },
   { "Sleep",           PTR_SLEEP },
   { "Sizing",          PTR_SIZING },
   { "SplitVertical",   PTR_SPLIT_VERTICAL },
   { "SplitHorizontal", PTR_SPLIT_HORIZONTAL },
   { "Magnifier",       PTR_MAGNIFIER },
   { "Hand",            PTR_HAND },
   { "HandLeft",        PTR_HAND_LEFT },
   { "HandRight",       PTR_HAND_RIGHT },
   { "Text",            PTR_TEXT },
   { "Paintbrush",      PTR_PAINTBRUSH },
   { "Stop",            PTR_STOP },
   { "Invisible",       PTR_INVISIBLE },
   { "Custom",          PTR_CUSTOM },
   { NULL, NULL }
};

static struct FieldDef RenderFlags[] = {
   { "INVISIBLE",     RNF_INVISIBLE   },   { "STICKTOBACK",   RNF_STICKTOBACK },
   { "STICKTOFRONT",  RNF_STICKTOFRONT },  { "VISIBLE",       RNF_VISIBLE },
   { "STICKY",        RNF_STICKY },        { "HOST",          RNF_HOST },
   { "REGION",        RNF_REGION },        { "FASTRESIZE",    RNF_FASTRESIZE },
   { "AUTOQUIT",      RNF_AUTOQUIT },      { "HASFOCUS",      RNF_HASFOCUS },
   { "VIDEO",         RNF_VIDEO },         { "ROOTLAYER",     RNF_ROOTLAYER },
   { "NOHORIZONTAL",  RNF_NOHORIZONTAL },  { "NOVERTICAL",    RNF_NOVERTICAL },
   { "SCROLLCONTENT", RNF_SCROLLCONTENT }, { "FIXEDBUFFER",   RNF_FIXEDBUFFER },
   { "PRECOPY",       RNF_PRECOPY },       { "AFTERCOPY",     RNF_AFTERCOPY },
   { "CUSTOMROOT",    RNF_CUSTOMROOT },    { "PERVASIVECOPY", RNF_PERVASIVECOPY },
   { "CURSOR",        RNF_CURSOR },        { "NOFOCUS",       RNF_NOFOCUS },
   { "TOTALREDRAW",   RNF_TOTALREDRAW },   { "POSTCOMPOSITE",   RNF_POSTCOMPOSITE },
   /*** Synonyms ***/
   { "MASK",          RNF_PRECOPY },
   { "TRANSPARENT",   RNF_INVISIBLE },
   { NULL, NULL }
};

static struct FieldArray RenderFields[] = {
   { "Drag",         0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "BitmapBuffer", 0, FDF_OBJECTID|FDF_R,   ID_BITMAP, NULL, NULL },
   { "Parent",       0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "RouteFocus",   0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, SET_RouteFocus },
   { "TopMargin",    0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "BottomMargin", 0, FDF_LONG|FDF_RW,      0, NULL, SET_BottomMargin },
   { "LeftMargin",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "RightMargin",  0, FDF_LONG|FDF_RW,      0, NULL, SET_RightMargin },
   { "MinWidth",     0, FDF_LONG|FDF_RW,      0, NULL, SET_MinWidth },
   { "MinHeight",    0, FDF_LONG|FDF_RW,      0, NULL, SET_MinHeight },
   { "MaxWidth",     0, FDF_LONG|FDF_RW,      0, NULL, SET_MaxWidth },
   { "MaxHeight",    0, FDF_LONG|FDF_RW,      0, NULL, SET_MaxHeight },
   { "LeftLimit",    0, FDF_LONG|FDF_RW,      0, NULL, SET_LeftLimit },
   { "RightLimit",   0, FDF_LONG|FDF_RW,      0, NULL, SET_RightLimit },
   { "TopLimit",     0, FDF_LONG|FDF_RW,      0, NULL, SET_TopLimit },
   { "BottomLimit",  0, FDF_LONG|FDF_RW,      0, NULL, SET_BottomLimit },
   { "Frame",        0, FDF_LONG|FDF_RW,      0, NULL, SET_Frame },
   { "Screen",       0, FDF_OBJECTID|FDF_R,   ID_SCREEN, NULL, NULL },
   { "Flags",        0, FDF_LONGFLAGS|FDF_RW, (LONG)&RenderFlags,   NULL, SET_Flags },
   { "XCoord",       0, FD_VARIABLE|FDF_LONG|FDF_PERCENTAGE|FDF_RW, 0, GET_XCoord,  SET_XCoord },
   { "YCoord",       0, FD_VARIABLE|FDF_LONG|FDF_PERCENTAGE|FDF_RW, 0, GET_YCoord,  SET_YCoord },
   { "Width",        0, FD_VARIABLE|FDF_LONG|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "Height",       0, FD_VARIABLE|FDF_LONG|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "RootLayer",    0, FDF_OBJECTID|FDF_RI,   0, NULL, NULL },
   { "TitleObject",  0, FDF_OBJECTID|FDF_RW,   0, NULL, NULL },
   { "Program",      0, FDF_LONG|FDF_RI,       0, NULL, NULL },
   { "Align",        0, FDF_LONGFLAGS|FDF_RW,  (LONG)&AlignFlags, NULL, NULL },
   { "Dimensions",   0, FDF_LONG|FDF_R,        0, NULL, NULL },
   { "KeyOutput",    0, FDF_OBJECTID|FDF_RI,   0, NULL, NULL },
   { "Cursor",       0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&CursorLookup, NULL, NULL },

   /*** Virtual fields ***/
   { "AbsX",          0, FDF_LONG|FDF_RW,     0,         GET_AbsX,           SET_AbsX },
   { "AbsY",          0, FDF_LONG|FDF_RW,     0,         GET_AbsY,           SET_AbsY },
   { "Bitmap",        0, FDF_OBJECT|FDF_R,    ID_BITMAP, GET_Bitmap,         NULL },
   { "BitsPerPixel",  0, FDF_LONG|FDF_RI,     0,         GET_BitsPerPixel,   SET_BitsPerPixel },
   { "Bottom",        0, FDF_LONG|FDF_R,      0,         GET_Bottom,         NULL },
   { "Buffer",        0, FDF_LONG|FDF_RI,     0,         GET_Buffer,         SET_Buffer },
   { "ColourRGB",     0, FDF_POINTER|FDF_RW,  0,         GET_ColourRGB,      SET_ColourRGB },
   { "Colour",        0, FDF_STRING|FDF_W,    0,         NULL,               SET_Colour },
   { "FastResize",    0, FDF_LONG|FDF_RI,     0,         GET_FastResize,     SET_FastResize },
   { "Hide",          0, FDF_LONG|FDF_INIT,   0,         NULL,               SET_Hide },
   { "InsideHeight",  0, FDF_LONG|FDF_RW,     0,         GET_InsideHeight,   SET_InsideHeight },
   { "InsideWidth",   0, FDF_LONG|FDF_RW,     0,         GET_InsideWidth,    SET_InsideWidth },
   { "Movement",      0, FDF_LONGFLAGS|FDF_RW,(LONG)&MovementFlags, NULL,    SET_Movement },
   { "Opacity",       0, FDF_FLOAT|FDF_RW,    0,         GET_Opacity,        SET_Opacity },
   { "Precopy",       0, FDF_LONG|FDF_RI,     0,         GET_Precopy,        SET_Precopy },
   { "Host",          0, FDF_LONG|FDF_RI,     0,         GET_Host,           SET_Host },
   { "Region",        0, FDF_LONG|FDF_RI,     0,         GET_Region,         SET_Region },
   { "Right",         0, FDF_LONG|FDF_R,      0,         GET_Right,          NULL },
   { "StickToBack",   0, FDF_LONG|FDF_W,      0,         NULL,               SET_StickToBack },
   { "StickToFront",  0, FDF_LONG|FDF_W,      0,         NULL,               SET_StickToFront },
   { "Sticky",        0, FDF_LONG|FDF_RW,     0,         GET_Sticky,         SET_Sticky },
   { "Invisible",     0, FDF_LONG|FDF_RI,     0,         GET_Invisible,      SET_Invisible },
   { "UserFocus",     0, FDF_LONG|FDF_R,      0,         GET_UserFocus,      NULL },
   { "Visible",       0, FDF_LONG|FDF_RW,     0,         GET_Visible,        SET_Visible },
   { "VisibleHeight", 0, FDF_LONG|FDF_R,      0,         GET_VisibleHeight,  NULL },
   { "VisibleWidth",  0, FDF_LONG|FDF_R,      0,         GET_VisibleWidth,   NULL },
   { "WindowHandle",  0, FDF_LONG|FDF_RW,     0,         GET_WindowHandle,   SET_WindowHandle },
   /*** Variable fields ***/
   { "X",             0, FDF_VARIABLE|FDF_LONG|FDF_PERCENTAGE|FDF_RW, 0, GET_XCoord,  SET_XCoord },
   { "XOffset",       0, FDF_VARIABLE|FDF_LONG|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",             0, FDF_VARIABLE|FDF_LONG|FDF_PERCENTAGE|FDF_RW, 0, GET_YCoord,  SET_YCoord },
   { "YOffset",       0, FDF_VARIABLE|FDF_LONG|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   /*** OBSOLETE - You can read these fields from GetRenderInfo(), or the Screen object referenced in the Screen field ***/
   { "AmtColours",    0, FDF_LONG|FDF_R,      0,         GET_AmtColours,     NULL },
   { "BytesPerPixel", 0, FDF_LONG|FDF_R,      0,         GET_BytesPerPixel,  NULL },
   { "Mask",          0, FDF_LONG|FDF_RI,     0,         GET_Precopy,        SET_Precopy },
   { "Transparent",   0, FDF_LONG|FDF_RI,     0,         GET_Invisible,      SET_Invisible },
   { "Root",          0, FDF_LONG|FDF_RI,     0,         GET_Host,           SET_Host },
   END_FIELD
};

static struct ActionArray RenderActions[] = {
   { AC_ActionNotify,     RENDER_ActionNotify },
   { AC_Activate,         RENDER_Activate },
   { AC_ClosingTag,       RENDER_ClosingTag },
   { AC_Disable,          RENDER_Disable },
   { AC_Draw,             RENDER_Draw },
   { AC_Enable,           RENDER_Enable },
   { AC_EventMessage,     RENDER_EventMessage },
   { AC_Focus,            RENDER_Focus },
   { AC_Free,             RENDER_Free },
   { AC_NewObject,        RENDER_NewObject },
   { AC_Hide,             RENDER_Hide },
   { AC_Init,             RENDER_Init },
   { AC_LostFocus,        RENDER_LostFocus },
   { AC_Move,             RENDER_Move },
   { AC_MoveToBack,       RENDER_MoveToBack },
   { AC_MoveToFront,      RENDER_MoveToFront },
   { AC_MoveToPoint,      RENDER_MoveToPoint },
   { AC_NewOwner,         RENDER_NewOwner },
   { AC_Redimension,      RENDER_Redimension },
   { AC_Resize,           RENDER_Resize },
   { AC_SaveImage,        RENDER_SaveImage },
   { AC_Scroll,           RENDER_Scroll },
   { AC_ScrollToPoint,    RENDER_ScrollToPoint },
   { AC_Show,             RENDER_Show },
   { AC_UserClick,        RENDER_UserClick },
   { AC_UserClickRelease, RENDER_UserClickRelease },
   { AC_UserMovement,     RENDER_UserMovement },
   { NULL, NULL }
};

/*****************************************************************************
** Class methods.
*/

static ERROR RENDER_Expose(struct Render *, struct mtExpose *);
static ERROR RENDER_InheritedFocus(struct Render *, struct mtInheritedFocus *);
static ERROR RENDER_InvalidateRegion(struct Render *, struct mtInvalidateRegion *);
static ERROR RENDER_MoveContent(struct Render *, struct mtMoveContent *);
static ERROR RENDER_SetCallback(struct Render *, struct mtSetRenderCallback *);
static ERROR RENDER_SetDisplay(struct Render *, struct mtSetRenderDisplay *);
static ERROR RENDER_SetOpacity(struct Render *, struct mtSetRenderOpacity *);

static struct FunctionField argsSetOpacity[]       = { { "Value", ARG_DOUBLE }, { "Adjustment", ARG_DOUBLE }, { NULL, NULL } };
static struct FunctionField argsInheritedFocus[]   = { { "Focus", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsMoveContent[]      = { { "XCoord", ARG_LONG }, { "YCoord", ARG_LONG }, { "Width", ARG_LONG }, { "Height", ARG_LONG }, { "XDest", ARG_LONG }, { "YDest", ARG_LONG }, { "ClipLeft", ARG_LONG }, { "ClipTop", ARG_LONG }, { "ClipRight", ARG_LONG }, { "ClipBottom", ARG_LONG }, { "Redraw", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsExpose[]           = { { "XCoord", ARG_LONG }, { "YCoord", ARG_LONG }, { "Width", ARG_LONG }, { "Height", ARG_LONG }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsInvalidateRegion[] = { { "XCoord", ARG_LONG }, { "YCoord", ARG_LONG }, { "Width", ARG_LONG }, { "Height", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsSetDisplay[]       = { { "XCoord", ARG_LONG }, { "YCoord", ARG_LONG }, { "Width", ARG_LONG }, { "Height", ARG_LONG }, { "InsideWidth", ARG_LONG }, { "InsideHeight", ARG_LONG }, { "BitsPerPixel", ARG_LONG }, { "RefreshRate", ARG_DOUBLE }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsSetCallback[]      = { { "Object", ARG_LONG }, { "Routine", ARG_PTR }, { NULL, NULL } };

static struct MethodArray RenderMethods[] = {
   { MT_SetRenderOpacity, RENDER_SetOpacity,       "SetOpacity",       argsSetOpacity,       sizeof(struct mtSetRenderOpacity) },
   { MT_Expose,           RENDER_Expose,           "Expose",           argsExpose,           sizeof(struct mtExpose) },
   { MT_InheritedFocus,   RENDER_InheritedFocus,   "InheritedFocus",   argsInheritedFocus,   sizeof(struct mtInheritedFocus) },
   { MT_MoveContent,      RENDER_MoveContent,      "MoveContent",      argsMoveContent,      sizeof(struct mtMoveContent) },
   { MT_InvalidateRegion, RENDER_InvalidateRegion, "InvalidateRegion", argsInvalidateRegion, sizeof(struct mtInvalidateRegion) },
   { MT_SetRenderDisplay, RENDER_SetDisplay,       "SetDisplay",       argsSetDisplay,       sizeof(struct mtSetRenderDisplay) },
   { MT_SetRenderCallback,RENDER_SetCallback,      "SetCallback",      argsSetCallback,      sizeof(struct mtSetRenderCallback) },
   { NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Internal: CreateRenderClass()
*/

static ERROR CreateRenderClass(void)
{
   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&RenderClass, NULL,
      FID_BaseClassID|TLONG, ID_RENDER,
      FID_Name|TSTRING,      "Render",
      FID_Category|TLONG,    CCF_GUI|CCF_DRAWABLE,
      FID_Actions|TPTR,      RenderActions,
      FID_Methods|TPTR,      RenderMethods,
      FID_Fields|TPTR,       RenderFields,
      FID_Size|TLONG,        sizeof(struct Render),
      FID_Version|TFLOAT,    VER_RENDER,
      FID_Flags|TLONG,       CLF_SHAREDOBJECTS|CLF_PUBLICOBJECTS,
      NULL));
}

/*****************************************************************************
** Render: ActionNotify
*/

static ERROR RENDER_ActionNotify(struct Render *Self, struct acActionNotify *NotifyArgs)
{
   if ((NotifyArgs->ActionID IS AC_Free) AND (NotifyArgs->Error IS ERR_Okay)) {
      if (NotifyArgs->ObjectID IS Self->ProgramID) {
         /*** Commit suicide if our linked task has disappeared ***/
         Action(AC_Free, Self, NULL);
      }
      else {
         /* Free ourselves in advance if our parent is in the process of being killed.
         ** This causes a chain reaction that results in a clean deallocation of
         ** the render hierarchy.
         */

         Self->Flags &= ~RNF_VISIBLE;
         UpdateRenderList(Self);

         acFree(Self);
      }
   }
   else if ((NotifyArgs->ActionID IS AC_Redimension) AND (NotifyArgs->Error IS ERR_Okay)) {
      struct RenderList *list;
      FLOAT parentwidth, parentheight, width, height, xcoord, ycoord;
      LONG i;

      /*** Get the width and height of our parent drawable ***/

      if ((list = AccessRenderList(ARF_READ, NULL))) {
         for (i=0; (list[i].ObjectID) AND (list[i].ObjectID != Self->ParentID); i++);
         if (!list[i].ObjectID) {
            ReleaseRenderList(ARF_READ);
            return(ObjectError(Self, ERH_ActionNotify, ERR_Search));
         }
         parentwidth  = list[i].Width;
         parentheight = list[i].Height;
         ReleaseRenderList(ARF_READ);
      }
      else return(ObjectError(Self, ERH_ActionNotify, ERR_AccessMemory));

      /*** Convert relative offsets to their fixed equivalent ***/

      if (Self->Dimensions & DMF_RELATIVE_XOFFSET) Self->prvXOffset = parentwidth * Self->prvXOffsetPercent / 100;
      if (Self->Dimensions & DMF_RELATIVE_YOFFSET) Self->prvYOffset = parentheight * Self->prvYOffsetPercent / 100;

      /*** Calculate absolute width and height values ***/

      if (Self->Dimensions & DMF_RELATIVE_WIDTH)   width = parentwidth * Self->prvWidthPercent / 100;
      else if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
      else if (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)) {
         if (Self->Dimensions & DMF_FIXED_X) {
            width = parentwidth - Self->XCoord - Self->prvXOffset;
         }
         else if (Self->Dimensions & DMF_RELATIVE_X) {
            width = parentwidth - (parentwidth * Self->prvXPercent / 100) - Self->prvXOffset;
         }
         else width = parentwidth - Self->prvXOffset;
      }
      else width = Self->Width;

      if (Self->Dimensions & DMF_RELATIVE_HEIGHT)   height = parentheight = Self->prvHeightPercent / 100;
      else if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
      else if (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)) {
         if (Self->Dimensions & DMF_FIXED_Y) {
            height = parentheight - Self->YCoord - Self->prvYOffset;
         }
         else if (Self->Dimensions & DMF_RELATIVE_Y) {
            height = parentheight - (parentheight * Self->prvYPercent / 100) - Self->prvYOffset;
         }
         else height = parentheight - Self->prvYOffset;
      }
      else height = Self->Height;

      /*** Calculate new coordinates ***/

      if (Self->Dimensions & DMF_RELATIVE_X) xcoord = (parentwidth * Self->prvXPercent / 100);
      else if (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)) xcoord = parentwidth - Self->prvXOffset - width;
      else xcoord = Self->XCoord;

      if (Self->Dimensions & DMF_RELATIVE_Y) ycoord = (parentheight * Self->prvYPercent / 100);
      else if (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)) ycoord = parentheight - Self->prvYOffset - height;
      else ycoord = Self->YCoord;

      /*** Alignment adjustments ***/

      if (Self->Align & ALIGN_LEFT) xcoord = 0;
      else if (Self->Align & ALIGN_RIGHT) xcoord = parentwidth - width;
      else if (Self->Align & ALIGN_HORIZONTAL) xcoord = (parentwidth - width) / 2;

      if (Self->Align & ALIGN_TOP) ycoord = 0;
      else if (Self->Align & ALIGN_BOTTOM) ycoord = parentheight - height;
      else if (Self->Align & ALIGN_VERTICAL) ycoord = (parentheight - height) / 2;

      /*** Perform the resize ***/

      ActionTags(AC_Redimension, Self, (DOUBLE)xcoord, (DOUBLE)ycoord, 0.0, (DOUBLE)width, (DOUBLE)height, 0.0);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DrawRegion()
*/

static void DrawRegion(OBJECTID RenderID, struct Render *Parent, struct Bitmap *Bitmap)
{
   struct Render *Self;
   LONG left, right, bottom, top, xoffset, yoffset;
   WORD i;

   if (AccessObject(RenderID, 5000, &Self) != ERR_Okay) return;

   /*** Only region objects can respond to draw messages ***/

   if (!(Self->Flags & (RNF_REGION|RNF_INVISIBLE))) {
      ReleaseObject(Self);
      return;
   }

   /*** If the render object is invisible, return immediately ***/

   if (!(Self->Flags & RNF_VISIBLE)) {
      ReleaseObject(Self);
      return;
   }

   if ((Self->Width < 1) OR (Self->Height < 1)) {
      ReleaseObject(Self);
      return;
   }

   if ((Self->XCoord > Bitmap->ClipRight) OR (Self->YCoord > Bitmap->ClipBottom) OR
       (Self->XCoord + Self->Width <= Bitmap->ClipLeft) OR
       (Self->YCoord + Self->Height <= Bitmap->ClipTop)) {
      ReleaseObject(Self);
      return;
   }

   /*** Take a copy of the current clipping and offset values ***/

   left    = Bitmap->ClipLeft;
   right   = Bitmap->ClipRight;
   bottom  = Bitmap->ClipBottom;
   top     = Bitmap->ClipTop;
   xoffset = Bitmap->XOffset;
   yoffset = Bitmap->YOffset;

   /*** Adjust clipping and offset values to match the absolute coordinates of our render object ***/

   Bitmap->XOffset += Self->XCoord;
   Bitmap->YOffset += Self->YCoord;

   /*** Adjust the clipping region of our parent so that it is relative to our rendered area ***/

   Bitmap->ClipLeft   -= Self->XCoord;
   Bitmap->ClipTop    -= Self->YCoord;
   Bitmap->ClipRight  -= Self->XCoord;
   Bitmap->ClipBottom -= Self->YCoord;

   /*** Make sure that the clipping values do not extend outside of our area ***/

   if (Bitmap->ClipLeft < 0) Bitmap->ClipLeft = 0;
   if (Bitmap->ClipTop < 0)  Bitmap->ClipTop = 0;
   if (Bitmap->ClipRight > Self->Width)   Bitmap->ClipRight = Self->Width;
   if (Bitmap->ClipBottom > Self->Height) Bitmap->ClipBottom = Self->Height;

   /*** Clear the Bitmap to the background colour if necessary ***/

   if (Self->Flags & RNF_CLEAR) {
      ActionTags(MT_DrawRGBRectangle, Bitmap, 0, 0,
         Self->Width, Self->Height, Self->prvColourRGB.Red,
         Self->prvColourRGB.Green, Self->prvColourRGB.Blue, TRUE);
   }

   /*** Call callbacks ***/

   if (Self->prvCallbackCount <= ARRAYSIZE(Self->prvCallbackCache)) {
      Self->prvCallback = Self->prvCallbackCache;
   }

   for (i=0; i < Self->prvCallbackCount; i++) {
      void (*routine)(LONG, struct Render *, struct Bitmap *);
      routine = Self->prvCallback[i].Routine;
      routine(Self->prvCallback[i].Data, Self, Bitmap);
   }

   /*** Draw the graphics to whatever Bitmap our parent is using ***/

   Self->prvDrawableBitmap = Bitmap;
   NotifySubscribers(Self, AC_Draw, NULL, NSF_LOCALTASK);
   Self->prvDrawableBitmap = NULL;

   Bitmap->ClipLeft   = left;
   Bitmap->ClipRight  = right;
   Bitmap->ClipBottom = bottom;
   Bitmap->ClipTop    = top;
   Bitmap->XOffset    = xoffset;
   Bitmap->YOffset    = yoffset;

   ReleaseObject(Self);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Shows a render object on the display.
###END###

*****************************************************************************/

static ERROR RENDER_Activate(struct Render *Self, APTR Void)
{
   if (!Self->ParentID) Action(AC_Show, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  ClosingTag
Short: Activates a render object.
###END###

*****************************************************************************/

static ERROR RENDER_ClosingTag(struct Render *Self, APTR Void)
{
   /* Show the render object unless the developer has
   ** specifically asked it to be hidden.
   */

   if (!(Self->Flags & RNF_HIDE)) {
      acShow(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Disables a render object.
###END###

*****************************************************************************/

static ERROR RENDER_Disable(struct Render *Self, APTR Void)
{
   Self->Flags |= RNF_DISABLED;
   UpdateRenderList(Self);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Enables a disabled render object.
###END###

*****************************************************************************/

static ERROR RENDER_Enable(struct Render *Self, APTR Void)
{
   Self->Flags &= ~RNF_DISABLED;
   UpdateRenderList(Self);
   return(ERR_Okay);
}

/*****************************************************************************
** Render: EventMessage
*/

static ERROR RENDER_EventMessage(struct Render *Self, struct acEventMessage *Args)
{
   struct RenderList *list;
   WORD i, level;
   UBYTE expose;

   if (Args->EventID IS EVT_TaskCrashed) {
      DPrintF("Render:","[Render:%d] Task crash detected.", Self->Head.UniqueID);

      if ((list = AccessRenderList(ARF_WRITE, NULL))) {

         for (i=0; (list[i].ObjectID) AND (list[i].ObjectID != Self->Head.UniqueID); i++);

         expose = FALSE;
         if (list[i].ObjectID) {
            level = list[i++].Level;
            while (list[i].Level > level) {
               if (CheckObjectExists(list[i].ObjectID, NULL) IS ERR_False) {
                  UntrackLayer(list[i].ObjectID);
                  expose = TRUE;
               }
               else i++;
            }
         }

         ReleaseRenderList(ARF_WRITE);

         if (expose) ExposeDrawable(Self->Head.UniqueID, 0, 0, Self->Width, Self->Height, EXF_CHILDREN);
      }
      else ObjectError(Self, AC_EventMessage, ERR_LockFailed);

      /*** If we are linked to the task that has crashed, commit suicide ***/

      if (Args->ObjectID IS Self->ProgramID) acFree(Self);
   }
   else if (Args->EventID IS EVT_UserLogin) {
      OBJECTPTR config;
      struct mtReadConfigInt read;
      struct mtReadConfigFloat readflt;
      LONG width, height, depth;
      DOUBLE gammared, gammagreen, gammablue, refreshrate;
      struct mtSetRenderDisplay display;
      struct mtSetGamma gamma;

      DPrintF("Render:","[Render:%d] New user login detected.", Self->Head.UniqueID);

      if (CreateObject(ID_CONFIG, NULL, (OBJECTPTR *)&config, NULL,
         FID_Location|TSTRING, "user:config/display.cfg",
         TAGEND) IS ERR_Okay) {

         refreshrate = -1.0;
         depth      = 32;
         gammared   = 1.0;
         gammagreen = 1.0;
         gammablue  = 1.0;
         width      = Self->Width;
         height     = Self->Height;

         read.Section = "SCREEN";
         read.Item = "Width";
         if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) width = read.Integer;
         read.Item = "Height";
         if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) height = read.Integer;
         read.Item = "Depth";
         if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) depth = read.Integer;

         readflt.Section = "SCREEN";
         readflt.Item = "RefreshRate";
         if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) refreshrate = readflt.Float;
         readflt.Item = "GammaRed";
         if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) gammared = readflt.Float;
         readflt.Item = "GammaGreen";
         if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) gammagreen = readflt.Float;
         readflt.Item = "GammaBlue";
         if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) gammablue = readflt.Float;

         if (width < 640) width = 640;
         if (height < 480) height = 480;

         display.XCoord       = 0;
         display.YCoord       = 0;
         display.Width        = width;
         display.Height       = height;
         display.InsideWidth  = display.Width;
         display.InsideHeight = display.Height;
         display.BitsPerPixel = depth;
         display.RefreshRate  = refreshrate;
         display.Flags        = NULL;
         Action(MT_SetRenderDisplay, Self, &display);

         gamma.Red   = gammared;
         gamma.Green = gammagreen;
         gamma.Blue  = gammablue;
         gamma.Flags = NULL;
         ActionMsg(MT_SetGamma, Self->ScreenID, &gamma);

         acFree(config);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Informs a render object that it has received the user focus.
###END###

*****************************************************************************/

static ERROR RENDER_Focus(struct Render *Self, APTR Void)
{
   struct RenderList *renderlist;
   struct ActionMessage *action;
   UBYTE msgbuffer[sizeof(struct Message) + sizeof(struct ActionMessage)];
   APTR queue;
   OBJECTID RenderID, oldlist[SIZE_FOCUSLIST], *focuslist;
   LONG pos, i, j, msgindex;
   WORD index;

   /* Delete back-logged focus messages (the last thing that we want
   ** is for our LostFocus action to execute only to be nullified by a
   ** Focus action that's been backlogged on the queue).
   */

   if (AccessMemory(GetResource(RES_MESSAGEQUEUE), MEM_READ, &queue) IS ERR_Okay) {
      msgindex = 0;
      while (ScanMessages(queue, &msgindex, MSGID_ACTION, msgbuffer, sizeof(msgbuffer)) IS ERR_Okay) {
         action = (struct ActionMessage *)(msgbuffer + sizeof(struct Message));

         if (action->ObjectID != Self->Head.UniqueID) continue;

         if ((action->ActionID IS AC_Focus) OR (action->ActionID IS AC_LostFocus) OR
             (action->ActionID IS MT_InheritedFocus)) {
            UpdateMessage(queue, ((struct Message *)msgbuffer)->UniqueID, -1, NULL, 0);
         }
      }
      ReleaseMemory(queue);
   }

   if (Self->Flags & RNF_NOFOCUS) return(ERR_Okay|ERF_Notified);

   //DPrintF("Focus()","[Render:%d] %s", Self->Head.UniqueID, Self->Head.Stats->Name);

   if (AccessMemory(RPM_FocusList, MEM_READWRITE, &focuslist) IS ERR_Okay) {
      /*** Return immediately if this render object already has the -primary- focus ***/

     if (focuslist[0] IS Self->Head.UniqueID) {
         ReleaseMemory(focuslist);
         return(ERR_Okay|ERF_Notified);
      }

      /*** Take a copy of the current focus list ***/

      for (i=0; focuslist[i]; i++) oldlist[i] = focuslist[i];
      oldlist[i] = NULL;

      /*** Build the new focus chain in a local focus list ***/

      if ((renderlist = AccessRenderList(ARF_READ, NULL))) {
         pos = 0;
         focuslist[pos++] = Self->Head.UniqueID; /* The primary focus tops the list */
         RenderID = Self->Head.UniqueID;

         /*** NOTE: This routine works but could be faster ***/

         while ((RenderID) AND (pos < SIZE_FOCUSLIST-1)) {
            for (index=0; (renderlist[index].ObjectID) AND (renderlist[index].ObjectID != RenderID); index++);

            if (renderlist[index].ObjectID IS RenderID) {
               if (renderlist[index].FocusID) {
                  RenderID = renderlist[index].FocusID;

                  /* Check if the routed focus loops back into the existing focus
                  ** chain (i.e. causes recursion).  If it does, use the parent render
                  ** object instead.
                  */

                  for (i=0; i < pos; i++) {
                     if (focuslist[i] IS RenderID) {
                        RenderID = renderlist[index].ParentID;
                        break;
                     }
                  }
               }
               else RenderID = renderlist[index].ParentID;
            }
            else break;

            focuslist[pos++] = RenderID;
         }
         focuslist[pos] = NULL;

         /*** Indicate which drawables have the focus in the renderlist ***/

         for (index=0; renderlist[index].ObjectID; index++) {
            renderlist[index].Flags &= ~RNF_HASFOCUS;
            for (pos=0; focuslist[pos]; pos++) {
               if (renderlist[index].ObjectID IS focuslist[pos]) {
                  renderlist[index].Flags |= RNF_HASFOCUS;
                  break;
               }
            }
         }

         ReleaseRenderList(ARF_READ);
      }
      else {
         ReleaseMemory(focuslist);
         return(SendError(ERH_Function, ERR_AccessMemory));
      }

      /* Send a Focus action to all parent render objects in our
      ** local focus list, on the condition that they do not
      ** already have the focus.
      */

      for (i=1; focuslist[i]; i++) {
         for (j=0; (oldlist[j]) AND (focuslist[i] != oldlist[j]); j++);
         if (focuslist[i] IS Self->Head.UniqueID) continue;
         if (focuslist[i] IS oldlist[j]) break; /* The new focus chain intersects with the previous chain, so we can discontinue the loop */
         ActionMsg(MT_InheritedFocus, focuslist[i], NULL);
      }

      /* Send out LostFocus actions to all objects that
      ** do not intersect with the new focus chain.
      */

      for (i=0; oldlist[i]; i++) {
         if (oldlist[i] IS Self->Head.UniqueID) continue;

         for (j=0; (focuslist[j]) AND (oldlist[i] != focuslist[j]); j++);
         if (!focuslist[j]) ActionMsg(AC_LostFocus, oldlist[i], NULL);
      }

      ReleaseMemory(focuslist);

      if (Self->Flags & RNF_HASFOCUS) {
         /*** Return without notification as we already have the focus ***/
         return(ERR_Okay|ERF_Notified);
      }
      else {
         if (Self->KeyOutputID) {
            OBJECTPTR keyboard, object;
            if (AccessObject(glKeyboardID, 2000, &keyboard) IS ERR_Okay) {
               DPrintF("Focus:","[Render] Subscribing object %d to keyboard %d.", Self->KeyOutputID, glKeyboardID);
               if (AccessObject(Self->KeyOutputID, 3000, &object) IS ERR_Okay) {
                  SubscribeChannel(keyboard, object);
                  ReleaseObject(object);
               }
               ReleaseObject(keyboard);
            }
            else DPrintF("@Focus:","[Render] Failed to access keyboard #%d.", glKeyboardID);
         }

         Self->Flags |= RNF_HASFOCUS;
         UpdateRenderList(Self);
         return(ERR_Okay);
      }
   }
   else return(ObjectError(Self, ERH_Focus, ERR_AccessMemory));
}

#if 0
/*** THIS IS THE OLD FOCUS CODE ***/

static ERROR RENDER_Focus(struct Render *Self, APTR Void)
{
   OBJECTID RenderID, list[SIZE_FOCUSLIST], *focuslist;
   struct RenderList *renderlist;
   LONG pos, i, j;
   WORD index;

   //DPrintF("Focus()","[Render:%d] %s", Self->Head.UniqueID, Self->Head.Stats->Name);

   if (AccessMemory(RPM_FocusList, MEM_READWRITE, &focuslist) IS ERR_Okay) {
      /*** Return immediately if this render object already has the primary focus ***/

      if (focuslist[0] IS Self->Head.UniqueID) {
         ReleaseMemory(focuslist);
         return(ERR_Okay|ERF_Notified);
      }

      /*** Scan the new focus chain into a local focus list ***/

      if ((renderlist = AccessRenderList(ARF_READ, NULL))) {
         pos = 0;
         list[pos++] = Self->Head.UniqueID;
         RenderID = Self->Head.UniqueID;

         while ((RenderID) AND (pos < SIZE_FOCUSLIST-1)) {
            for (index=0; (renderlist[index].ObjectID) AND (renderlist[index].ObjectID != RenderID); index++);

            if (renderlist[index].ObjectID IS RenderID) {
               if (renderlist[index].FocusID) {
                  RenderID = renderlist[index].FocusID;

                  /* Check if the routed focus loops back into the existing focus chain.
                  ** If it does, use the parent render object instead.
                  */

                  for (i=0; i < pos; i++) {
                     if (list[i] IS RenderID) {
                        RenderID = renderlist[index].ParentID;
                        break;
                     }
                  }
               }
               else RenderID = renderlist[index].ParentID;
            }
            else break;

            list[pos++] = RenderID;
         }
         list[pos] = NULL;

         ReleaseRenderList(ARF_READ);
      }
      else {
         ReleaseMemory(focuslist);
         return(SendError(ERH_Function, ERR_AccessMemory));
      }

      /* Send a Focus action to all parent render objects in our
      ** local focus list, on the condition that they do not
      ** already have the focus.
      */

      for (i=0; list[i]; i++) {
         for (j=0; (focuslist[j]) AND (list[i] != focuslist[j]); j++);
         if (list[i] != Self->Head.UniqueID) {
            if (list[i] IS focuslist[j]) break;
            else ActionMsg(MT_InheritedFocus, list[i], NULL);
         }
      }

      /* Send out LostFocus actions to all objects that
      ** do not intersect with the new focus chain.
      */

      for (i=0; focuslist[i]; i++) {
         if (focuslist[i] IS Self->Head.UniqueID) continue;

         for (j=0; (list[j]) AND (focuslist[i] != list[j]); j++);
         if (!list[j]) ActionMsg(AC_LostFocus, focuslist[i], NULL);
      }

      /*** Copy the new focus list over the old focus list ***/

      for (i=0; list[i]; i++) focuslist[i] = list[i];
      focuslist[i] = NULL;

      ReleaseMemory(focuslist);

      if (Self->Flags & RNF_HASFOCUS) {
         /*** Return without notification as we already have the focus ***/
         return(ERR_Okay|ERF_Notified);
      }
      else {
         Self->Flags |= RNF_HASFOCUS;
         UpdateRenderList(Self);
         return(ERR_Okay);
      }
   }
   else return(ObjectError(Self, ERH_Focus, ERR_AccessMemory));
}
#endif

static ERROR RENDER_InheritedFocus(struct Render *Self, struct mtInheritedFocus *Args)
{
   if (Self->Flags & RNF_HASFOCUS) {
      return(ERR_Okay);
   }
   else {
      Self->Flags |= RNF_HASFOCUS;

      //UpdateRenderList(Self); /* Not necessary because RENDER_Focus sets the renderlist */

      if (Self->KeyOutputID) {
         OBJECTPTR keyboard, object;
         if (AccessObject(glKeyboardID, 2000, &keyboard) IS ERR_Okay) {
            if (AccessObject(Self->KeyOutputID, 3000, &object) IS ERR_Okay) {
               SubscribeChannel(keyboard, object);
               ReleaseObject(object);
            }
            ReleaseObject(keyboard);
         }
      }

      NotifySubscribers(Self, AC_Focus, NULL, NULL);
      return(ERR_Okay);
   }
}

/*****************************************************************************
** Render: Free
*/

static ERROR RENDER_Free(struct Render *Self, APTR Void)
{
   struct Render *parent;
   OBJECTPTR broadcast;
   ERROR error;

   if (Self->prvButtonFlags) {
      /* If there are unreleased button flags on this object, we have
      ** to fake a UserClickRelease call in case there are subscribers
      ** that require notification of mouse buttons are being dropped.
      */
      struct acUserClickRelease release;

      release.XCoord = 0;
      release.YCoord = 0;
      release.ZCoord = 0;
      release.XChange = 0;
      release.YChange = 0;
      release.ZChange = 0;
      release.OverObject = Self->Head.UniqueID;
      release.ButtonFlags = Self->prvButtonFlags;
      Action(AC_UserClickRelease, Self, &release);
   }

   if (Self->prvPointerWithin) {
      /* If user-movement has been recently detected on the drawable
      ** but not released at the time of termination, send a fake
      ** UserMovement call to balance the ledger.
      */

      struct RenderList *list;
      struct acUserMovement usermove;

      if ((list = AccessRenderList(ARF_READ, NULL))) {
         usermove.OverObject = list[0].ObjectID;
         ReleaseRenderList(ARF_READ);
      }
      else usermove.OverObject = NULL;

      usermove.XCoord  = -1;
      usermove.YCoord  = -1;
      usermove.ZCoord  = 0;
      usermove.XChange = 0;
      usermove.YChange = 0;
      usermove.ZChange = 0;
      Action(AC_UserMovement, Self, &usermove);
   }

   if (!Self->ParentID) {
      if (AccessObject(glBroadcastID, 5000, &broadcast) IS ERR_Okay) {
         mtUnsubscribeEvent(broadcast, Self->Head.UniqueID, EVT_TaskCrashed);
         mtUnsubscribeEvent(broadcast, Self->Head.UniqueID, EVT_UserLogin);
         ReleaseObject(broadcast);
      }
   }

   if (Self->prvCallbackCount > ARRAYSIZE(Self->prvCallbackCache)) {
      FreeMemory(Self->prvCallback);
      Self->prvCallback = NULL;
   }

   if (Self->ParentID) {
      if ((error = AccessObject(Self->ParentID, 5000, &parent)) IS ERR_Okay) {
         UnsubscribeAction(parent, NULL, GetUniqueID(Self));
         if (Self->Flags & (RNF_REGION|RNF_INVISIBLE)) ActionTags(MT_SetRenderCallback, parent, Self->Head.UniqueID, NULL);
         ReleaseObject(parent);
      }
   }

   acHide(Self);

   /*** Remove any references to this render object from the global render list ***/

   UntrackLayer(Self->Head.UniqueID);

   /*** Unsubscribe from the parent ***/

   if ((!Self->ParentID) AND (Self->ScreenID)) {
      acFreeID(Self->ScreenID);
      Self->ScreenID = NULL;
   }

   if ((Self->BufferID) AND ((!Self->prvBitmapOwnerID) OR (Self->prvBitmapOwnerID IS Self->Head.UniqueID))) {
      if (Self->prvBitmap) {
         ReleaseObject(Self->prvBitmap);
         Self->prvBitmap = NULL;
      }
      acFreeID(Self->BufferID);
      Self->BufferID = NULL;
   }

   /* Give the focus to the parent if our object has the primary focus.  Do not
   ** apply this technique to render objects acting as windows, as the window class
   ** has its own focus management code.
   */

   if ((Self->Flags & RNF_HASFOCUS) AND (Self->prvWindow IS FALSE)) {
      if (Self->ParentID) {
         acFocusID(Self->ParentID);
      }
   }

   if ((Self->Flags & RNF_HASFOCUS) AND (Self->KeyOutputID)) {
      OBJECTPTR keyboard;
      if (AccessObject(glKeyboardID, 2000, &keyboard) IS ERR_Okay) {
         UnsubscribeChannel(keyboard, Self->KeyOutputID);
         ReleaseObject(keyboard);
      }
   }

   if (Self->Flags & RNF_AUTOQUIT) {
      DPrintF("Free:","[Render:%d] Posting a quit message due to use of AUTOQUIT.", Self->Head.UniqueID);
      if ((Self->Head.TaskID IS Self->ProgramID) OR (!Self->ProgramID)) {
         SendMessage(NULL, MSGID_QUIT, NULL, NULL, NULL);
      }
      else {
         struct ListTasks *list;
         WORD i, message;

         message = FALSE;
         if (ListTasks(NULL, &list) IS ERR_Okay) {
            for (i=0; list[i].TaskID; i++) {
               if (list[i].TaskID IS Self->ProgramID) {
                  SendMessage(list[i].MessageID, MSGID_QUIT, NULL, NULL, NULL);
                  message = TRUE;
                  break;
               }
            }
            FreeMemory(list);
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Hides a render object from the display.
###END###

*****************************************************************************/

static ERROR RENDER_Hide(struct Render *Self, APTR Void)
{
   #ifdef DBG_DRAW
      DPrintF("~Hide()","[Render:%d]", Self->Head.UniqueID);
   #endif

   if (!(Self->Flags & RNF_VISIBLE)) {
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return(ERR_Okay|ERF_Notified);
   }

   if (!Self->ParentID) {
      if (acHideID(Self->ScreenID) != ERR_Okay) {
         #ifdef DBG_DRAW
            StepBack();
         #endif
         return(ERR_Failed);
      }

      Self->Flags &= ~RNF_VISIBLE;
      UpdateRenderList(Self);
   }
   else {
      /* Mark this render object as invisible, then invalidate the region
      ** it was covering in order to have the background redrawn.
      */

      Self->Flags &= ~RNF_VISIBLE;
      UpdateRenderList(Self);

      if (Self->Flags & RNF_REGION) {
         RedrawDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, IRF_RELATIVE);
         ExposeDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, NULL);
      }
      else {
         if (Self->prvBitmapOwnerID != Self->Head.UniqueID) {
            RedrawDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, IRF_RELATIVE);
         }
         ExposeDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, EXF_CHILDREN|EXF_REDRAW_VOLATILE);
      }
   }

   #ifdef DBG_DRAW
      StepBack();
   #endif
   return(ERR_Okay);
}

/*****************************************************************************
** Render: Init
*/

static ERROR RENDER_Init(struct Render *Self, APTR Void)
{
   struct MemInfo info;
   struct Screen *screen;
   struct Render *parent;
   struct Bitmap *bitmap;
   ERROR error;
   OBJECTID parent_bitmap, bitmap_owner;
   LONG memflags, scrflags;
   BYTE require_store;

   require_store = FALSE;
   parent_bitmap = NULL;
   bitmap_owner  = NULL;

   if (GetClassID(Self->Head.OwnerID) IS ID_WINDOW) Self->prvWindow = TRUE;

   if (!Self->RootLayer) Self->RootLayer = Self->Head.UniqueID;
   else if (Self->RootLayer != Self->Head.UniqueID) Self->Flags |= RNF_CUSTOMROOT;

   if (Self->Flags & RNF_CURSOR) Self->Flags |= RNF_STICKTOFRONT;

   if (Self->ParentID) {
      if (AccessObject(Self->ParentID, 3000, &parent) != ERR_Okay) {
         return(ObjectError(Self, ERH_Init, ERR_AccessObject));
      }

      error = ERR_Okay;

      /* Check that the parent is either part of our task space,
      ** or has the root option set in any other case.
      */

      if (parent->Head.TaskID != CurrentTaskID()) {
         if (!(parent->Flags & RNF_HOST)) {
            DPrintF("!Init:","[Render:%d] Cannot initialise to parent #%d as it has not set the host drawable option.", Self->Head.UniqueID, Self->ParentID);
            ReleaseObject(parent);
            return(ERR_Failed);
         }
      }

      if (Self->Flags & RNF_REGION) {
         /* Regions must share the same task space with their parent.  If we
         ** can't meet this requirement, turn off the region flag.
         */

         if (parent->Head.TaskID != CurrentTaskID()) {
            DPrintF("@Init:","[Render:%d] Region cannot initialise to parent #%d - not in our task space.", Self->Head.UniqueID, Self->ParentID);
            Self->Flags &= ~RNF_REGION;
         }
      }

      /* If the parent drawable is a region, the child must also be a region
      ** or our drawing system will get confused.
      */

      if ((parent->Flags & RNF_REGION) AND (!(Self->Flags & RNF_REGION))) {
         Self->Flags |= RNF_REGION;
      }

      /* If the parent has the ROOTLAYER flag set, we have to inherit
      ** whatever root layer that the parent is using, as well as the PRECOPY
      ** and/or AFTERCOPY and opacity flags if they are set.
      */

      if (parent->Flags & RNF_ROOTLAYER) {
         Self->Flags |= RNF_ROOTLAYER; // | (parent->Flags & (RNF_PRECOPY|RNF_AFTERCOPY)); precopy/aftercopy bits not necessary? - interferes with fading anyway...
         if (Self->RootLayer IS Self->Head.UniqueID) Self->RootLayer = parent->RootLayer;
      }

      /*** Subscribe to the render parent's Resize and Redimension actions ***/

      SubscribeActionTags(parent, Self,
         AC_Free,
         AC_Redimension,
         TAGEND);

      /* If the render object is a simple region, subscribe
      ** to the Draw action of the parent object.
      */

      if (Self->Flags & RNF_REGION) {
         ActionTags(MT_SetRenderCallback, parent, Self->Head.UniqueID, &DrawRegion);

         /* Turn off flags that should never be combined with regions.
         */

         if (Self->Flags & RNF_PRECOPY) Self->Flags &= ~RNF_CLEAR;

         Self->Flags &= ~(RNF_INVISIBLE|RNF_AFTERCOPY|RNF_POSTCOMPOSITE|RNF_HOST);
      }
      else if (Self->Flags & RNF_INVISIBLE) {
         ActionTags(MT_SetRenderCallback, parent, Self->Head.UniqueID, &DrawRegion);
         Self->Flags &= ~(RNF_REGION|RNF_PRECOPY|RNF_AFTERCOPY|RNF_POSTCOMPOSITE|RNF_CLEAR);
      }

      /*** Recalculate coordinates if offsets are used ***/

      if (Self->Dimensions & DMF_FIXED_XOFFSET)    SetField(Self, FID_XOffset, FD_LONG, &Self->prvXOffset);
      if (Self->Dimensions & DMF_FIXED_YOFFSET)    SetField(Self, FID_YOffset, FD_LONG, &Self->prvYOffset);
      if (Self->Dimensions & DMF_RELATIVE_XOFFSET) SetField(Self, FID_XOffset, FD_FLOAT|FD_PERCENTAGE, &Self->prvXOffsetPercent);
      if (Self->Dimensions & DMF_RELATIVE_YOFFSET) SetField(Self, FID_YOffset, FD_FLOAT|FD_PERCENTAGE, &Self->prvYOffsetPercent);
      if (Self->Dimensions & DMF_RELATIVE_X)       SetField(Self, FID_XCoord,  FD_FLOAT|FD_PERCENTAGE, &Self->prvXPercent);
      if (Self->Dimensions & DMF_RELATIVE_Y)       SetField(Self, FID_YCoord,  FD_FLOAT|FD_PERCENTAGE, &Self->prvYPercent);
      if (Self->Dimensions & DMF_RELATIVE_WIDTH)   SetField(Self, FID_Width,   FD_FLOAT|FD_PERCENTAGE, &Self->prvWidthPercent);
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT)  SetField(Self, FID_Height,  FD_FLOAT|FD_PERCENTAGE, &Self->prvHeightPercent);

      if (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & (DMF_RELATIVE_XOFFSET|DMF_FIXED_XOFFSET)) {
            Self->Width = parent->Width - Self->XCoord - Self->prvXOffset;
         }
         else {
            Self->Width = 20;
            Self->Dimensions |= DMF_FIXED_WIDTH;
         }
      }

      if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & (DMF_RELATIVE_YOFFSET|DMF_FIXED_YOFFSET)) {
            Self->Height = parent->Height - Self->YCoord - Self->prvYOffset;
         }
         else {
            Self->Height = 20;
            Self->Dimensions |= DMF_FIXED_HEIGHT;
         }
      }

      /*** Alignment adjustments ***/

      if (Self->Align & ALIGN_LEFT) { Self->XCoord = 0; SetField(Self, FID_XCoord, FD_LONG, &Self->XCoord); }
      else if (Self->Align & ALIGN_RIGHT) { Self->XCoord = parent->Width - Self->Width; SetField(Self, FID_XCoord, FD_LONG, &Self->XCoord); }
      else if (Self->Align & ALIGN_HORIZONTAL) { Self->XCoord = (parent->Width - Self->Width) / 2; SetField(Self, FID_XCoord, FD_LONG, &Self->XCoord); }

      if (Self->Align & ALIGN_TOP) { Self->YCoord = 0; SetField(Self, FID_YCoord, FD_LONG, &Self->YCoord); }
      else if (Self->Align & ALIGN_BOTTOM) { Self->YCoord = parent->Height - Self->Height; SetField(Self, FID_YCoord, FD_LONG, &Self->YCoord); }
      else if (Self->Align & ALIGN_VERTICAL) { Self->YCoord = (parent->Height - Self->Height) / 2; SetField(Self, FID_YCoord, FD_LONG, &Self->YCoord); }

      if (Self->Height < Self->MinHeight + Self->TopMargin  + Self->BottomMargin) Self->Height = Self->MinHeight + Self->TopMargin  + Self->BottomMargin;
      if (Self->Width  < Self->MinWidth  + Self->LeftMargin + Self->RightMargin)  Self->Width  = Self->MinWidth  + Self->LeftMargin + Self->RightMargin;
      if (Self->Height > Self->MaxHeight + Self->TopMargin  + Self->BottomMargin) Self->Height = Self->MaxHeight + Self->TopMargin  + Self->BottomMargin;
      if (Self->Width  > Self->MaxWidth  + Self->LeftMargin + Self->RightMargin)  Self->Width  = Self->MaxWidth  + Self->LeftMargin + Self->RightMargin;

      Self->ScreenID        = parent->ScreenID;
      Self->prvScreenWindow = parent->prvScreenWindow;
      parent_bitmap         = parent->BufferID;
      bitmap_owner          = parent->prvBitmapOwnerID;

      /* Check if the parent is a host object - if so then we will
      ** need a backing store so that we can interact correctly
      ** with foreign processes that jump in at the same layer
      ** position.  This does not apply to regions, of course.
      */

      if ((parent->Flags & RNF_HOST) AND (!(Self->Flags & (RNF_REGION|RNF_INVISIBLE)))) require_store = TRUE;

      ReleaseObject(parent);
   }
   else {
      OBJECTPTR config;
      struct mtReadConfigInt read;
      struct mtReadConfigFloat readflt;
      LONG depth;
      DOUBLE gammared, gammagreen, gammablue, refreshrate;

      DPrintF("Init:","[Render:%d] This render object will be screen-based.", Self->Head.UniqueID);

      /*** Turn off any flags that may not be used for the top-most layer ***/

      Self->Flags &= ~(RNF_REGION|RNF_INVISIBLE|RNF_PRECOPY|RNF_POSTCOMPOSITE|RNF_AFTERCOPY);

      refreshrate = -1.0;
      depth = 32;
      gammared   = 1.0;
      gammagreen = 1.0;
      gammablue  = 1.0;
      scrflags = NULL;
      if (CreateObject(ID_CONFIG, NULL, (OBJECTPTR *)&config, NULL,
         FID_Location|TSTRING, "user:config/display.cfg",
         TAGEND) IS ERR_Okay) {
         WORD display;

         display = ScreenDisplayType();

         read.Section = "SCREEN";

         if (display IS DT_WINDOWS) {
            read.Item = "Maximise";
            if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) {
               if (read.Integer) scrflags |= SCR_MAXIMISE;
            }
         }

         if ((display IS DT_X11) OR (display IS DT_WINDOWS)) {
            if (!Self->Width) {
               read.Item = "WindowWidth";
               if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) {
                  Self->Width = read.Integer;
               }
               else {
                  read.Item = "Width";
                  if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) {
                     Self->Width = read.Integer;
                  }
               }
            }

            if (!Self->Height) {
               read.Item = "WindowHeight";
               if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) {
                  Self->Height = read.Integer;
               }
               else {
                  read.Item = "Height";
                  if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) {
                     Self->Height = read.Integer;
                  }
               }
            }

            read.Item = "WindowX";
            if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) {
               Self->XCoord = read.Integer;
            }

            read.Item = "WindowY";
            if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) {
               Self->YCoord = read.Integer;
            }
         }
         else {
            if (!Self->Width) {
               read.Item = "Width";
               if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) Self->Width = read.Integer;
            }

            if (!Self->Height) {
               read.Item = "Height";
               if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) Self->Height = read.Integer;
            }

            if (!Self->XCoord) {
               read.Item = "XCoord";
               if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) Self->XCoord = read.Integer;
            }

            if (!Self->YCoord) {
               read.Item = "YCoord";
               if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) Self->YCoord = read.Integer;
            }

            read.Item = "Depth";
            if (Action(MT_ReadConfigInt, config, &read) IS ERR_Okay) depth = read.Integer;
         }

         readflt.Section = "SCREEN";
         readflt.Item = "RefreshRate";
         if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) refreshrate = readflt.Float;

         readflt.Item = "GammaRed";
         if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) gammared = readflt.Float;

         readflt.Item = "GammaGreen";
         if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) gammagreen = readflt.Float;

         readflt.Item = "GammaBlue";
         if (Action(MT_ReadConfigFloat, config, &readflt) IS ERR_Okay) gammablue = readflt.Float;

         acFree(config);
      }

      if (ScreenDisplayType() IS DT_WINDOWS) {
         Self->LeftMargin   = 4;
         Self->RightMargin  = 4;
         Self->TopMargin    = 4;
         Self->BottomMargin = 4;
      }

      if ((Self->Width < 10) OR (Self->Height < 6)) {
         Self->Width = 640;
         Self->Height = 480;
      }

      if (Self->Height < Self->MinHeight + Self->TopMargin  + Self->BottomMargin) Self->Height = Self->MinHeight + Self->TopMargin  + Self->BottomMargin;
      if (Self->Width  < Self->MinWidth  + Self->LeftMargin + Self->RightMargin)  Self->Width  = Self->MinWidth  + Self->LeftMargin + Self->RightMargin;
      if (Self->Height > Self->MaxHeight + Self->TopMargin  + Self->BottomMargin) Self->Height = Self->MaxHeight + Self->TopMargin  + Self->BottomMargin;
      if (Self->Width  > Self->MaxWidth  + Self->LeftMargin + Self->RightMargin)  Self->Width  = Self->MaxWidth  + Self->LeftMargin + Self->RightMargin;

      if (CreateObject(ID_SCREEN, NF_CHILD|NF_PUBLIC|NF_SHARED|Self->Head.ObjectFlags, &screen, &Self->ScreenID,
            FID_Name|TSTRING,        "SystemScreen",
            FID_XCoord|TLONG,        Self->XCoord,
            FID_YCoord|TLONG,        Self->YCoord,
            FID_Width|TLONG,         Self->Width,
            FID_Height|TLONG,        Self->Height,
            FID_BitsPerPixel|TLONG,  depth,
            FID_RefreshRate|TDOUBLE, refreshrate,
            FID_GammaRed|TDOUBLE,    gammared,
            FID_GammaGreen|TDOUBLE,  gammagreen,
            FID_GammaBlue|TDOUBLE,   gammablue,
            FID_Flags|TLONG,         scrflags,
            FID_WindowHandle|TLONG,  Self->prvScreenWindow, /* Sometimes a window may be preset, e.g. for a web plugin */
            TAGEND) IS ERR_Okay) {

         Self->Width  = screen->Width;
         Self->Height = screen->Height;
         Self->XCoord = 0;
         Self->YCoord = 0;

         struct mtScreenSizeHints hints;

         if ((Self->MaxWidth) OR (Self->MaxHeight) OR (Self->MinWidth) OR (Self->MinHeight)) {
            if (Self->MaxWidth > 0)  hints.MaxWidth  = Self->MaxWidth  + Self->LeftMargin + Self->RightMargin;  else hints.MaxWidth  = 0;
            if (Self->MaxHeight > 0) hints.MaxHeight = Self->MaxHeight + Self->TopMargin  + Self->BottomMargin; else hints.MaxHeight = 0;
            if (Self->MinWidth > 0)  hints.MinWidth  = Self->MinWidth  + Self->LeftMargin + Self->RightMargin;  else hints.MinWidth  = 0;
            if (Self->MinHeight > 0) hints.MinHeight = Self->MinHeight + Self->TopMargin  + Self->BottomMargin; else hints.MinHeight = 0;
            Action(MT_ScreenSizeHints, screen, &hints);
         }

         acFlush(screen);

         /* For hosted environments, record the window handle (NB:
         ** this is doubling up the screen handle, we should just
         ** make the window handle a virtual field so that we don't
         ** need a permanent record of it).
         */

         GetField(screen, FID_WindowHandle, FT_LONG, &Self->prvScreenWindow);

         #ifdef _WIN32
            winSetRenderID(Self->prvScreenWindow, Self->Head.UniqueID);
         #endif

         ReleaseObject(screen);
      }
      else return(ObjectError(Self, ERH_Init, ERR_CreateObject));
   }

   /* Allocate a backing store if this is a host object, OR the parent
   ** is foreign, OR we are the child of a host object (check made earlier),
   ** OR render object is masked.
   */

   if (!Self->ParentID) require_store = TRUE;

   if (Self->Flags & (RNF_HOST|RNF_PRECOPY|RNF_AFTERCOPY|RNF_CURSOR)) require_store = TRUE;

   if ((require_store IS FALSE) AND (Self->ParentID)) {
      if (MemoryIDInfo(Self->ParentID, VER_MemoryInfo, &info) IS ERR_Okay) {
         if (info.TaskID != CurrentTaskID()) require_store = TRUE;
      }
   }

   if (Self->Flags & (RNF_REGION|RNF_INVISIBLE)) require_store = FALSE;

   if (require_store) {
      Self->prvBitmapOwnerID = Self->Head.UniqueID;

      if (AccessObject(Self->ScreenID, 3000, &screen) IS ERR_Okay) {
         LONG bpp;

         memflags = MEM_DATA;

         if (Self->Flags & RNF_VIDEO) {
            /* If acceleration is available then it is OK to
            ** create the buffer in video RAM.
            */
            if (!(screen->Flags & SCR_NOACCELERATION)) memflags = MEM_BLIT;
         }

         if (Self->Flags & RNF_POSTCOMPOSITE) {
            /*** If dynamic compositing will be used then we must have an alpha channel ***/
            bpp = 32;
         }
         else if (Self->prvBitsPerPixel) {
            bpp = Self->prvBitsPerPixel; /* BPP has been preset by the user */
            DPrintF("Init:","[Render] Preset depth of %d bpp detected.", bpp);
         }
         else bpp = screen->Bitmap->BitsPerPixel;

         if (CreateObject(ID_BITMAP, NF_CHILD|NF_PUBLIC|NF_SHARED|Self->Head.ObjectFlags, &bitmap, &Self->BufferID,
            FID_BitsPerPixel|TLONG, bpp,
            FID_Width|TLONG,        Self->Width,
            FID_Height|TLONG,       Self->Height,
            FID_DataFlags|TLONG,    memflags,
            FID_Flags|TLONG,        (Self->Flags & RNF_POSTCOMPOSITE) ? (BMF_ALPHA|BMF_FIXEDDEPTH) : NULL,
            TAGEND) IS ERR_Okay) {

            Self->prvBitsPerPixel  = bitmap->BitsPerPixel;
            Self->prvBytesPerPixel = bitmap->BytesPerPixel;
            Self->prvLineWidth     = bitmap->LineWidth;
            Self->prvDataMID       = bitmap->DataMID;
            ReleaseObject(bitmap);
         }
         else {
            ReleaseObject(screen);
            return(ObjectError(Self, ERH_Init, ERR_CreateObject));
         }
         ReleaseObject(screen);
      }
      else return(ObjectError(Self, ERH_Init, ERR_AccessObject));
   }
   else {
      Self->BufferID         = parent_bitmap;
      Self->prvBitmapOwnerID = bitmap_owner;
   }

   /*** If the FIXEDBUFFER option is set, pass the NEVERSHRINK option to the bitmap ***/

   if (Self->Flags & RNF_FIXEDBUFFER) {
      if (AccessObject(Self->BufferID, 5000, &bitmap) IS ERR_Okay) {
         bitmap->Flags |= BMF_NEVERSHRINK;
         ReleaseObject(bitmap);
      }
   }

   /*** Track the render object ***/

   if (TrackLayer(Self) != ERR_Okay) return(ERR_Failed);

   /* Move the render object to the back of the render list when
   ** stick-to-back is enforced.
   */

   if (Self->Flags & RNF_STICKTOBACK) Action(AC_MoveToBack, Self, NULL);

   /* Listen to the TaskCrashed event if we are a host render object.
   ** This will allow us to clean up the RenderList when a task crashes.
   ** Listening to the UserLogin event also allows us to switch to the user's
   ** preferred display format on login.
   */

   if ((!Self->ParentID) AND (StrCompare("SystemDisplay", Self->Head.Stats->Name, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
      OBJECTPTR broadcast;
      if (AccessObject(glBroadcastID, 5000, &broadcast) IS ERR_Okay) {
         mtSubscribeEvent(broadcast, Self->Head.UniqueID, EVT_TaskCrashed);
         mtSubscribeEvent(broadcast, Self->Head.UniqueID, EVT_UserLogin);
         ReleaseObject(broadcast);
      }
   }

   if (!Self->ProgramID) Self->ProgramID = Self->Head.TaskID;
   else if (Self->ProgramID != Self->Head.TaskID) {
      OBJECTPTR task;
      if (AccessObject(Self->ProgramID, 4000, &task) IS ERR_Okay) {
         SubscribeActionTags(task, Self, AC_Free, TAGEND);
         ReleaseObject(task);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Informs a render object that it has lost the user focus.
###END###

*****************************************************************************/

static ERROR RENDER_LostFocus(struct Render *Self, APTR Void)
{
   struct ActionMessage *action;
   UBYTE msgbuffer[sizeof(struct Message) + sizeof(struct ActionMessage)];
   LONG msgindex;
   APTR queue;

   /* Delete backlogged focus messages (the last thing that we want
   ** is for our LostFocus action to execute only to be nullified by a
   ** Focus action that's been backlogged on the queue).
   */

   if (AccessMemory(GetResource(RES_MESSAGEQUEUE), MEM_READ, &queue) IS ERR_Okay) {
      msgindex = 0;
      while (ScanMessages(queue, &msgindex, MSGID_ACTION, msgbuffer, sizeof(msgbuffer)) IS ERR_Okay) {
         action = (struct ActionMessage *)(msgbuffer + sizeof(struct Message));

         if (action->ObjectID != Self->Head.UniqueID) continue;

         if ((action->ActionID IS AC_Focus) OR (action->ActionID IS AC_LostFocus) OR
             (action->ActionID IS MT_InheritedFocus)) {
            UpdateMessage(queue, ((struct Message *)msgbuffer)->UniqueID, -1, NULL, 0);
         }
      }
      ReleaseMemory(queue);
   }

   /*** Drop the focus ***/

   Self->Flags &= ~RNF_HASFOCUS;
   UpdateRenderList(Self);

   if (Self->KeyOutputID) {
      OBJECTPTR keyboard;
      if (AccessObject(glKeyboardID, 2000, &keyboard) IS ERR_Okay) {
         UnsubscribeChannel(keyboard, Self->KeyOutputID);
         ReleaseObject(keyboard);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves a render object to the back of its container.
###END###

*****************************************************************************/

static ERROR RENDER_MoveToBack(struct Render *Self, APTR Void)
{
   struct RenderList *list;
   OBJECTID parent_bitmap;
   WORD pos, index, total, i, level;

   if (!Self->ParentID) return(ERR_Okay|ERF_Notified);

   DPrintF("~MoveToBack()","[Render:%d] %s", Self->Head.UniqueID, Self->Head.Stats->Name);

   if ((list = AccessRenderList(ARF_WRITE, NULL))) {
      /*** Get our position within the chain ***/

      for (index=0; (list[index].ObjectID) AND (list[index].ObjectID != Self->Head.UniqueID); index++);

      if (list[index].ObjectID != Self->Head.UniqueID) {
         ReleaseRenderList(ARF_WRITE);
         StepBack();
         return(ObjectError(Self, ERH_MoveToBack, ERR_Search)|ERF_Notified);
      }

      if ((i = FindDrawableIndex(list, list[index].ParentID)) != -1) {
         parent_bitmap = list[i].BitmapID;
      }
      else parent_bitmap = 0;

      /*** Find the position in the list that our render object will be moved to ***/

      pos = index;
      level = list[index].Level;
      for (i=index-1; (i >= 0) AND (list[i].Level >= level); i--) {
         if (list[i].Level IS level) {

            if (Self->prvBitmapOwnerID IS Self->Head.UniqueID) {
               /* If we own an independent bitmap, we cannot move behind
               ** drawables that are members of the parent region
               */
               if (list[i].BitmapID IS parent_bitmap) break;
            }

            if (!(Self->Flags & RNF_STICKTOBACK) AND (list[i].Flags & RNF_STICKTOBACK)) break;
            pos = i;
         }
      }

      /*** If the position is unchanged, return immediately ***/

      if (pos >= index) {
         ReleaseRenderList(ARF_WRITE);
         StepBack();
         return(ERR_Okay|ERF_Notified);
      }

      /* Reorder the list so that our render object
      ** is inserted at the new position
      */

      for (total=1; list[index+total].Level > list[index].Level; total++);

      {
         struct RenderList tmp[total];
         CopyMemory(list + index, &tmp, sizeof(struct RenderList) * total); /* Copy the source entry into a buffer */
         CopyMemory(list + pos, list + pos + total, sizeof(struct RenderList) * (index - pos)); /* Shift everything behind us to the front */
         CopyMemory(&tmp, list + pos, sizeof(struct RenderList) * total); /* Copy our source entry to its new position */
      }

      if (Self->Flags & RNF_VISIBLE) {
         /*** Redraw our background if we are volatile ***/

         if (CheckVolatile(list, index)) RedrawDrawable(Self->Head.UniqueID, list[pos].AbsX, list[pos].AbsY, list[pos].Right, list[pos].Bottom, NULL);

         /*** Expose changes to the display ***/

         ExposeDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, EXF_CHILDREN|EXF_REDRAW_VOLATILE_OVERLAP);
      }

      ReleaseRenderList(ARF_WRITE);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToFront
Short: Moves a render object to the front of its container.
###END###

*****************************************************************************/

static ERROR RENDER_MoveToFront(struct Render *Self, APTR Void)
{
   struct RenderList *list;
   OBJECTPTR parent;
   WORD index, currentindex, total, i, level;

   DPrintF("~MoveToFront()","[Render:%d] %s", Self->Head.UniqueID, Self->Head.Stats->Name);

   if (!Self->ParentID) {
      StepBack();
      return(ERR_Okay|ERF_Notified);
   }

   if ((list = AccessRenderList(ARF_WRITE, NULL))) {
      /*** Get our position within the chain ***/

      for (currentindex=0; (list[currentindex].ObjectID) AND (list[currentindex].ObjectID != Self->Head.UniqueID); currentindex++);

      if (list[currentindex].ObjectID != Self->Head.UniqueID) {
         DPrintF("@MoveToFront:","[Render:%d] I am not registered in the global render list.", Self->Head.UniqueID);
         ReleaseRenderList(ARF_WRITE);
         StepBack();
         return(ERR_Search|ERF_Notified);
      }

      /*** Find the object in the list that our render object will displace ***/

      index = currentindex;
      level = list[currentindex].Level;
      for (i=currentindex+1; (list[i].Level >= list[currentindex].Level); i++) {
         if (list[i].Level IS level) {
            if (list[i].Flags & RNF_POINTER) break;

            if (Self->prvBitmapOwnerID != Self->Head.UniqueID) {
               /* If we are a member of our parent's bitmap, we cannot be moved in front
               ** of bitmaps that own an independent buffer.
               */
               if (list[i].BitmapID != Self->BufferID) break;
            }

            if (!(Self->Flags & RNF_STICKTOFRONT) AND (list[i].Flags & RNF_STICKTOFRONT)) break;
            index = i;
         }
      }

      /* If the render object is a region, resubscribe to the Draw action to
      ** move our drawable region to the front of the subscription list.
      */

      if (Self->Flags & RNF_REGION) {
         if (AccessObject(Self->ParentID, 3000, &parent) IS ERR_Okay) {
            ActionTags(MT_SetRenderCallback, parent, Self->Head.UniqueID, &DrawRegion);
            ReleaseObject(parent);
         }
      }

      /*** If the position hasn't changed, return immediately ***/

      if (index <= currentindex) {
         ReleaseRenderList(ARF_WRITE);
         StepBack();
         return(ERR_Okay|ERF_Notified);
      }

      /*** Skip past the children that belong to the target object ***/

      level = list[index].Level;
      while (list[index+1].Level > level) index++;

      /* Count the number of children that have been
      ** assigned to our render object.
      */

      for (total=1; list[currentindex+total].Level > list[currentindex].Level; total++);

      /* Reorder the list so that our render object
      ** is inserted at the new index.
      */

      {
         struct RenderList tmp[total];
         CopyMemory(list + currentindex, &tmp, sizeof(struct RenderList) * total); /* Copy the source entry into a buffer */
         CopyMemory(list + currentindex + total, list + currentindex, sizeof(struct RenderList) * (index - currentindex - total + 1)); /* Shift everything in front of us to the back */
         index = index - total + 1;
         CopyMemory(&tmp, list + index, sizeof(struct RenderList) * total); /* Copy our source entry to its new index */
      }

      if (Self->Flags & RNF_VISIBLE) {
         struct Bitmap *bitmap;

         /* A redraw is required for:
         **   Any volatile regions that were in front of our drawable prior to the move-to-front
         **     (by moving to the front, their background has been changed).
         **   Areas of our drawable that were obscured by drawables that also shared our bitmap space.
         */

         if (AccessObject(Self->BufferID, 5000, &bitmap) IS ERR_Okay) {
            InvalidateExposedOverlap(Self, list, currentindex, index, list[index].AbsX, list[index].AbsY,
               list[index].Right, list[index].Bottom, bitmap);
            ReleaseObject(bitmap);
         }

         /*if (Self->Flags & RNF_VOLATILE) ExposeDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, EXF_CHILDREN|EXF_REDRAW_VOLATILE_OVERLAP);
         else*/
         ExposeDrawable(Self->Head.UniqueID, 0, 0, Self->Width, Self->Height, EXF_CHILDREN|EXF_REDRAW_VOLATILE_OVERLAP);

         /* This redraw method is extremely lazy/slow compared to the above, because it redraws
         ** the entire drawable and all preceding drawables.  However, it is guaranteed to work
         ** in all cases.
         */
         /*
         InvalidateRegion(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, NULL);
         if (Self->Flags & RNF_VOLATILE) ExposeDrawable(Self->Head.UniqueID, 0, 0, Self->Width, Self->Height, EXF_CHILDREN);
         else ExposeDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, EXF_CHILDREN);
         */
      }

      ReleaseRenderList(ARF_WRITE);

      StepBack();
      return(ERR_Okay);
   }
   else {
      StepBack();
      return(ObjectError(Self, ERH_MoveToFront, ERR_AccessMemory)|ERF_Notified);
   }
}

/*****************************************************************************
** Internal: InvalidateExposedOverlap()
**
** This function will expose areas that are uncovered when a drawable changes
** its position in the rendering tree (e.g. moving towards the front).
**
** This function is only interested in siblings of the drawable that we've moved.
** Also, any intersecting drawables need to share the same bitmap surface.
**
** All coordinates are expressed in absolute format.
*/

static void InvalidateExposedOverlap(struct Render *Self, struct RenderList *list, LONG OldIndex, LONG Index,
   LONG Left, LONG Top, LONG Right, LONG Bottom, struct Bitmap *Bitmap)
{
   LONG listx, listy, listright, listbottom;
   WORD i, j, level;

#ifdef DBG_DRAW
   DPrintF("~InvalidateOverlap()","[Render:%d] %dx%d %dx%d", Self->Head.UniqueID, Left, Top, Right-Left, Bottom-Top);
#endif

   if ((list[Index].Flags & (RNF_REGION|RNF_INVISIBLE)) OR (!(list[Index].Flags & RNF_VISIBLE))) {
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return;
   }

   level = list[Index].Level;

   for (i=OldIndex; i < Index; i++) {
      /* A redraw is required for:
      **   Any volatile regions that were in front of our drawable prior to the move-to-front (by moving to the front, their background has been changed).
      **   Areas of our drawable that were obscured by drawables that also shared our bitmap space.
      */

      if (!(list[i].Flags & RNF_VISIBLE)) goto skipcontent;
      if (list[i].Flags & RNF_REGION) goto skipcontent;
      if (list[i].Flags & RNF_INVISIBLE) continue;

      if (list[i].BitmapID != list[Index].BitmapID) {
         if (list[i].Flags & RNF_VOLATILE) {
            /* The drawable is volatile and on a different bitmap - it will have to be redrawn
            ** because its background has changed.  It will not have to be exposed because our
            ** drawable is sitting on top of it.
            */

            RedrawDrawable(list[i].ObjectID, Left, Top, Right, Bottom, NULL);
         }
         else if (!(Self->Flags & RNF_VOLATILE)) goto skipcontent; /* Even though this drawable is on a different bitmap, we cannot ignore it when we are volatile */
      }

      if ((list[i].Left < Right) AND (list[i].Top < Bottom) AND
          (list[i].Right > Left) AND (list[i].Bottom > Top)) {
         /* Intersecting drawable discovered.  What we do now is keep
         ** scanning for other overlapping siblings to restrict our
         ** exposure space (so that we don't repeat expose drawing for
         ** overlapping areas).  Then we call RedrawDrawable() to draw the exposed
         ** area.
         */

         listx      = list[i].Left;
         listy      = list[i].Top;
         listright  = list[i].Right;
         listbottom = list[i].Bottom;

         if (Left > listx)        listx      = Left;
         if (Top > listy)         listy      = Top;
         if (Bottom < listbottom) listbottom = Bottom;
         if (Right < listright)   listright  = Right;

         RedrawDrawable(Self->Head.UniqueID, listx, listy, listright, listbottom, NULL);
      }

skipcontent:
      /*** Skip past any children of the overlapping object ***/

      for (j=i+1; list[j].Level > list[i].Level; j++);
      i = j - 1;
   }

#ifdef DBG_DRAW
   StepBack();
#endif
}

/*****************************************************************************

###ACTION###
Name:  MoveToPoint
Short: Moves a render object to an absolute coordinate.
###END###

*****************************************************************************/

static ERROR RENDER_MoveToPoint(struct Render *Self, struct acMoveToPoint *Args)
{
   struct acMove move;

   if (Args->Flags & MTF_X) move.XChange = Args->XCoord - Self->XCoord;
   else move.XChange = 0;

   if (Args->Flags & MTF_Y) move.YChange = Args->YCoord - Self->YCoord;
   else move.YChange = 0;

   move.ZChange = 0;

   Action(AC_Move, Self, &move);

   return(ERR_Okay);
}

/*****************************************************************************
** Render: NewOwner()
*/

static ERROR RENDER_NewOwner(struct Render *Self, struct acNewOwner *Args)
{
   OBJECTID OwnerID;

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) {
      OwnerID = Args->NewOwnerID;
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (OwnerID) Self->ParentID = OwnerID;
      else Self->ParentID = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Render: NewObject
*/

static ERROR RENDER_NewObject(struct Render *Self, APTR Void)
{
   /*** Set defaults ***/
   Self->LeftLimit   = -1000000000;
   Self->RightLimit  = -1000000000;
   Self->TopLimit    = -1000000000;
   Self->BottomLimit = -1000000000;
   Self->MaxWidth    = 20000;
   Self->MaxHeight   = 20000;
   Self->MinWidth    = 1;
   Self->MinHeight   = 1;
   Self->Frame       = 1;
   Self->prvOpacity  = 255;
   Self->RootLayer   = Self->Head.UniqueID;
   Self->ProgramID   = Self->Head.TaskID;
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  SaveImage
Short: Saves the graphical image of a render object.

###DESCRIPTION###
<p>If you need to store the image (graphical content) of a render object,
use the SaveImage action.  Calling SaveImage on a render object will cause it
to generate an image of its contents and save them to the given destination
object.  Any child drawables in the region will also be included in the
resulting image data.</p>

<p>The image data will be saved in the data format that is indicated by the
setting in the ClassID argument.  Options are limited to members of the
Picture class, for example ID_PNG, ID_JPEG, ID_PCX and ID_PICTURE.  If no
ClassID is specified, the default file format is used.</p>

###END###

*****************************************************************************/

static ERROR RENDER_SaveImage(struct Render *Self, struct acSaveImage *Args)
{
   struct Picture *picture;
   struct Bitmap *bitmap;
   struct RenderList *list;
   CLASSID ClassID;
   OBJECTID bitmapid;
   WORD i, j, level;

   if (!Args) return(ObjectError(Self, ERH_SaveImage, ERR_Args));

   DPrintF("~SaveImage:","[Render:%d]", Self->Head.UniqueID);

   /*** Create a Bitmap that is the same size as the rendered area ***/

   if (!Args->ClassID) ClassID = ID_PICTURE;
   else ClassID = Args->ClassID;

   if (NewPrivateObject(ClassID, NULL, &picture) IS ERR_Okay) {
      picture->Flags |= IMG_NEW;
      picture->Bitmap->Width  = Self->Width;
      picture->Bitmap->Height = Self->Height;

      if ((bitmap = AccessVideo(Self->ScreenID))) {
         picture->Bitmap->BitsPerPixel  = bitmap->BitsPerPixel;
         picture->Bitmap->BytesPerPixel = bitmap->BytesPerPixel;
         picture->Bitmap->Type          = bitmap->Type;
         ReleaseVideo();
      }

      if (acInit(picture) IS ERR_Okay) {

         /*** Scan through the render list and blit each buffer to our picture ***/

         if ((list = AccessRenderList(ARF_READ, NULL))) {

            for (i=0; (list[i].ObjectID != Self->Head.UniqueID) AND (list[i].ObjectID); i++);

            bitmapid = NULL;
            for (j=i; (list[j].ObjectID) AND ((j IS i) OR (list[j].Level > list[i].Level)); j++) {
               if ((!(list[j].Flags & RNF_VISIBLE)) OR (list[j].Flags & RNF_CURSOR)) {
                  /*** Skip this drawable area and all invisible children ***/
                  level = list[j].Level;
                  while (list[j+1].Level > level) j++;
                  continue;
               }

               /*** If the bitmaps are different, we have found something new to blit ***/

               if (list[j].BitmapID != bitmapid) {

                  bitmapid = list[j].BitmapID;

                  if (list[j].Flags & RNF_REGION) continue;

                  BlitDrawable(list[j].ObjectID, picture->Bitmap, NULL, 0, 0,
                     list[j].Width, list[j].Height,
                     list[j].AbsX - list[i].AbsX, list[j].AbsY - list[i].AbsY);
               }
            }

            ReleaseRenderList(ARF_READ);

            /*** Save the picture to disk ***/

            if (Action(AC_SaveImage, picture, Args) IS ERR_Okay) {
               acFree(picture);
               StepBack();
               return(ERR_Okay);
            }
         }
      }

      acFree(picture);
      return(StepObjectError(Self, ERH_SaveImage, ERR_Failed));
   }
   else return(StepObjectError(Self, ERH_SaveImage, ERR_NewObject));
}

/*****************************************************************************

###ACTION###
Name:  Scroll
Short: Scrolls rendered content to a new position.

###DESCRIPTION###
<p>Calling the Scroll action on a render object with the SCROLLCONTENT flag
set will cause it to move its contents in the requested direction.  The Render class uses the Move action
to achieve scrolling, so any objects that do not support Move will remain at
their given position.  Everything else will be shifted by the same amount of
units as specified in the XChange, YChange and ZChange arguments.</p>

<p>Some objects may support a 'sticky' field that can be set to TRUE to
prevent them from being moved.  This feature is present in the Render class,
amongst others.</p>

<p>If the render object does not have the SCROLLCONTENT flag set, the call
will flow through to any objects that may be listening for the Scroll
action on the drawable.</p>

###END###

*****************************************************************************/

static ERROR RENDER_Scroll(struct Render *Self, struct acScroll *Args)
{
   struct RenderList *list;
   struct acMove move;
   WORD i, level;

   if (!Args) return(ERR_Args);

   if (Self->Flags & RNF_SCROLLCONTENT) {
      if ((Args->XChange >= 1) OR (Args->XChange <= -1) OR
          (Args->YChange >= 1) OR (Args->YChange <= -1)) {
         move.XChange = -Args->XChange;
         move.YChange = -Args->YChange;
         move.ZChange = -Args->ZChange;

         if ((list = AccessRenderList(ARF_READ, NULL))) {
            /*** Find our object ***/

            for (i=0; list[i].ObjectID; i++) {
               if (Self->Head.UniqueID IS list[i].ObjectID) {
                  /*** Send a move command to each child drawable ***/
                  level = list[i].Level + 1;
                  for (++i; list[i].Level >= level; i++) {
                     if (list[i].Level IS level) DelayMsg(AC_Move, list[i].ObjectID, &move);
                  }
                  break;
               }
            }

            ReleaseRenderList(ARF_READ);
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  ScrollToPoint
Short: Moves the content of a render object to a specific point.
###END###

*****************************************************************************/

static ERROR RENDER_ScrollToPoint(struct Render *Self, struct acScrollToPoint *Args)
{
   struct RenderList *list;
   struct acMoveToPoint move;
   WORD i, level;

   if (!Args) return(ERR_Args);

   if (Self->Flags & RNF_SCROLLCONTENT) {
      if (Args->Flags & STF_X) move.XCoord = -Args->XCoord;
      else move.XCoord = 0;

      if (Args->Flags & STF_Y) move.YCoord = -Args->YCoord;
      else move.YCoord = 0;

      if (Args->Flags & STF_Z) move.ZCoord = -Args->ZCoord;
      else move.ZCoord = 0;

      move.Flags = Args->Flags;

      if ((list = AccessRenderList(ARF_READ, NULL))) {
         /*** Find our object ***/

         for (i=0; list[i].ObjectID; i++) {
            if (Self->Head.UniqueID IS list[i].ObjectID) {
               level = list[i].Level + 1;
               for (++i; list[i].Level >= level; i++) {
                  if (list[i].Level IS level) DelayMsg(AC_MoveToPoint, list[i].ObjectID, &move);
               }
               break;
            }
         }

         ReleaseRenderList(ARF_READ);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    SetRenderCallback
Synonym: SetCallback
Short:   Allows you to hook into the drawing process of a rendered object.

###DESCRIPTION###
<p>The SetRenderCallback method is used to draw graphics to render areas.  By
hooking into the render object you will be able to draw to its Bitmap every
time that the Draw action is executed.  The render object will call the
address specified in the Routine argument using the following function
specification:</p>

<pre>
   void Draw(LONG Data, struct Render *Drawable, struct Bitmap *Bitmap)
</pre>

<p>The Data argument is a customised 32-bit parameter that you specify when
calling the SetRenderCallback method for the first time.  The Drawable
argument refers to the address of the render object that you subscribed to.
The Bitmap argument refers to the bitmap object that your graphics must be
drawn to.</p>

<p>To draw to the render object, write to the bitmap surface as you would
with any bitmap object (refer to the Bitmap class).  To get the width and
height of the available drawing space, please read the Width and Height
fields from the Drawable object.</p>

###INPUT###
LONG/Data: 32-bit data to be passed as the first argument to the draw function.
APTR/Routine: Pointer to the callback routine.

###RESULT###
ERR_Okay: Success
ERR_Args: Invalid arguments were specified.

###SEE ALSO###
Class: Bitmap

###END###

*****************************************************************************/

static ERROR RENDER_SetCallback(struct Render *Self, struct mtSetRenderCallback *Args)
{
   WORD i;

   if ((!Args) OR (!Args->Data)) return(ObjectError(Self, MT_SetRenderCallback, ERR_Args));

   DPrintF("SetCallback()","[%d] Data: %d, Routine: $%.8x (Count: %d)", Self->Head.UniqueID, Args->Data, Args->Routine, Self->prvCallbackCount);

   if (Self->prvCallbackCount <= ARRAYSIZE(Self->prvCallbackCache)) {
      /*** Ensure that the prvCallback address is valid ***/
      Self->prvCallback = Self->prvCallbackCache;
   }

   if (!Args->Routine) {
      /*** Clear an existing routine from the callback list ***/

      if (!Self->prvCallback) return(ERR_Search);

      for (i=0; (i < Self->prvCallbackCount) AND (Self->prvCallback[i].Data != Args->Data); i++);

      if (i >= Self->prvCallbackCount) return(ERR_Search);

      Self->prvCallback[i].Data = 0;
      Self->prvCallback[i].Routine = 0;
      while (i < Self->prvCallbackCount-1) {
         Self->prvCallback[i].Data    = Self->prvCallback[i+1].Data;
         Self->prvCallback[i].Routine = Self->prvCallback[i+1].Routine;
         i++;
      }

      Self->prvCallbackCount--;

      if (Self->prvCallbackCount IS ARRAYSIZE(Self->prvCallbackCache)) {
         for (i=0; i < ARRAYSIZE(Self->prvCallbackCache); i++) {
            Self->prvCallbackCache[i].Data    = Self->prvCallback[i].Data;
            Self->prvCallbackCache[i].Routine = Self->prvCallback[i].Routine;
         }
         FreeMemory(Self->prvCallback);
         Self->prvCallback = Self->prvCallbackCache;
      }

      return(ERR_Okay);
   }
   else {
      /*** Check if the subscription is already on the list for our render object ***/

      if (Self->prvCallback) {
         for (i=0; (i < Self->prvCallbackCount) AND (Self->prvCallback[i].Data != Args->Data); i++);
      }
      else i = Self->prvCallbackCount;

      if (i < Self->prvCallbackCount) {
         /*** Re-subscribe the data/routine by moving it to the 'foreground' of the draw-list ***/

         while (i < Self->prvCallbackCount-1) {
            Self->prvCallback[i].Data    = Self->prvCallback[i+1].Data;
            Self->prvCallback[i].Routine = Self->prvCallback[i+1].Routine;
            i++;
         }
         Self->prvCallback[i].Data    = Args->Data;
         Self->prvCallback[i].Routine = Args->Routine;
      }
      else if (Self->prvCallbackCount < ARRAYSIZE(Self->prvCallbackCache)) {
         /*** Add the callback routine to the cache ***/

         Self->prvCallback = Self->prvCallbackCache;
         Self->prvCallback[Self->prvCallbackCount].Data    = Args->Data;
         Self->prvCallback[Self->prvCallbackCount].Routine = Args->Routine;
         Self->prvCallbackCount++;
      }
      else {
         /*** The fast cache is unavailable - add the routine to the memory list ***/

         if (Self->prvCallbackCount IS ARRAYSIZE(Self->prvCallbackCache)) {
            if (AllocMemory(sizeof(struct RenderCallback) * (Self->prvCallbackCount + 1), MEM_DATA, &Self->prvCallback, NULL) IS ERR_Okay) {
               for (i=0; i < ARRAYSIZE(Self->prvCallbackCache); i++) {
                  Self->prvCallback[i].Data    = Self->prvCallbackCache[i].Data;
                  Self->prvCallback[i].Routine = Self->prvCallbackCache[i].Routine;
               }
               Self->prvCallback[Self->prvCallbackCount].Data    = Args->Data;
               Self->prvCallback[Self->prvCallbackCount].Routine = Args->Routine;
               Self->prvCallbackCount++;
            }
            else return(ObjectError(Self, MT_SetRenderCallback, ERR_AllocMemory));
         }
         else if (ReallocMemory(Self->prvCallback, sizeof(struct RenderCallback) * (Self->prvCallbackCount + 1), &Self->prvCallback, NULL) IS ERR_Okay) {
            Self->prvCallback[Self->prvCallbackCount].Data    = Args->Data;
            Self->prvCallback[Self->prvCallbackCount].Routine = Args->Routine;
            Self->prvCallbackCount++;
         }
         else return(ObjectError(Self, MT_SetRenderCallback, ERR_ReallocMemory));
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    SetOpacity
Synonym: SetRenderOpacity
Short:   Alters the opacity of a render object.

###DESCRIPTION###
<p>This method will change the opacity of the drawable and execute a redraw
to make the changes to the display.</p>

###INPUT###
DOUBLE/Value: The new opacity value between 0 and 100% (ignored if you have set the Adjustment parameter).
DOUBLE/Adjustment: Adjustment value to add or subtract from the existing opacity (set to zero if you want to set a fixed Value instead).

###RESULT###
ERR_Okay: The opacity of the render object was changed.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR RENDER_SetOpacity(struct Render *Self, struct mtSetRenderOpacity *Args)
{
   FLOAT value;

   if (!Args) return(ObjectError(Self, ERH_Method, ERR_Args));

   if (Self->prvBitmapOwnerID != Self->Head.UniqueID) {
      DPrintF("@SetOpacity:","[Render] Opacity cannot be set on a drawable that does not own its bitmap.");
      return(ERR_NoSupport);
   }

   if (Args->Adjustment) {
      value = (Self->prvOpacity * 100 / 255) + Args->Adjustment;
      SET_Opacity(Self, &value);
   }
   else {
      value = Args->Value;
      SET_Opacity(Self, &value);
   }

   /* Use the DelayMsg() feature so that we don't end up with major
   ** lag problems when SetOpacity is being used for things like
   ** fading.
   */

   if (Self->Flags & RNF_VISIBLE) DelayMsg(MT_InvalidateRegion, Self->Head.UniqueID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Shows a render object on the display.
###END###

*****************************************************************************/

static ERROR RENDER_Show(struct Render *Self, APTR Void)
{
   LONG notified;

   if (Self->Flags & RNF_VISIBLE) {
      DPrintF("~6Show()","[Render:%d]", GetUniqueID(Self));
      notified = ERF_Notified;
   }
   else {
      DPrintF("~2Show()","[Render:%d]", GetUniqueID(Self));
      notified = NULL;
   }

   if (!Self->ParentID) {
      if (acShowID(Self->ScreenID) IS ERR_Okay) {
         Self->Flags |= RNF_VISIBLE;
      }
      else {
         StepBack();
         return(ObjectError(Self, ERH_Show, ERR_Failed));
      }
   }
   else Self->Flags |= RNF_VISIBLE;

   UpdateRenderList(Self);

   if (Self->Flags & RNF_REGION) {
      RedrawDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, IRF_RELATIVE);
      ExposeDrawable(Self->ParentID, Self->XCoord, Self->YCoord, Self->Width, Self->Height, NULL);
   }
   else {
      RedrawDrawable(Self->Head.UniqueID, 0, 0, Self->Width, Self->Height, IRF_RELATIVE);
      ExposeDrawable(Self->Head.UniqueID, 0, 0, Self->Width, Self->Height, EXF_CHILDREN|EXF_REDRAW_VOLATILE_OVERLAP);
   }

   StepBack();
   return(ERR_Okay|notified);
}

/*****************************************************************************

###ACTION###
Name:  UserClick
Short: Informs a render object that the user has clicked on it.
###END###

*****************************************************************************/

static ERROR RENDER_UserClick(struct Render *Self, struct acUserClick *Args)
{
   struct RenderList *list;
   struct Render *drawable;
   WORD index, dragindex;
   BYTE anchored;

   if (Self->Flags & RNF_DISABLED) return(ERR_Okay|ERF_Notified);
   Self->prvButtonFlags |= Args->ButtonFlags;

   /*** Send a focus message to our object ***/

   if (!(Self->Flags & RNF_NOFOCUS)) acFocus(Self);

   /*** Anchor the pointer position if dragging is enabled ***/

   if ((Self->DragID) AND (Self->prvButtonFlags & JD_LMB)) {
      #ifdef DBG_DRAW
         DPrintF("UserClick:","[%d] Dragging object %d", Self->Head.UniqueID, Self->DragID);
      #endif

      anchored = DRAG_NORMAL;
      if (SetCursor(Self->DragID, ACF_ANCHOR|ACF_LMB, (Self->CursorID) ? Self->CursorID : PTR_DEFAULT, NULL, Self->Head.UniqueID, NULL) IS ERR_Okay) {
         if (ScreenDisplayType() IS DT_NATIVE) anchored = DRAG_ANCHOR;
      }

      if (Self->DragID IS Self->Head.UniqueID) {
         Self->prvAnchored = anchored;
         Self->prvAnchorX  = Args->XCoord;
         Self->prvAnchorY  = Args->YCoord;
      }
      else if (AccessObject(Self->DragID, 3000, &drawable) IS ERR_Okay) {
         if ((list = AccessRenderList(ARF_READ, NULL))) {
            dragindex = FindDrawableIndex(list, Self->DragID);
            index     = FindDrawableIndex(list, Self->Head.UniqueID);

            drawable->prvAnchored = anchored;
            drawable->prvAnchorX  = (list[index].Left + Args->XCoord) - list[dragindex].Left;
            drawable->prvAnchorY  = (list[index].Top + Args->YCoord) - list[dragindex].Top;
            ReleaseRenderList(ARF_READ);
         }
         else RestoreCursor(NULL, Self->Head.UniqueID); /* Undo changes */

         ReleaseObject(drawable);
      }
      else RestoreCursor(NULL, Self->Head.UniqueID); /* Undo changes */
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserClickRelease
Short: Used to inform a render object when a user click has been released.
###END###

*****************************************************************************/

static ERROR RENDER_UserClickRelease(struct Render *Self, struct acUserClickRelease *Args)
{
   struct Render *drawable;

   if ((Self->Flags & RNF_DISABLED) AND (!Self->prvButtonFlags)) return(ERR_Okay|ERF_Notified);
   Self->prvButtonFlags = (~Args->ButtonFlags) & Self->prvButtonFlags;

   if ((!(Self->prvButtonFlags & JD_LMB)) AND ((Self->DragID) OR (Self->prvAnchored))) {
      Self->prvAnchored = DRAG_NONE;

      if (Self->DragID != Self->Head.UniqueID) {
         if (AccessObject(Self->DragID, 3000, &drawable) IS ERR_Okay) {
            drawable->prvAnchored = DRAG_NONE;
            ReleaseObject(drawable);
         }
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  UserMovement
Short: Used to notify render object's of user movement in and out of their graphical areas.
###END###

*****************************************************************************/

static ERROR RENDER_UserMovement(struct Render *Self, struct acUserMovement *Args)
{
   UBYTE buffer[sizeof(struct Message) + sizeof(struct ActionMessage) + sizeof(struct acUserMovement)];
   struct ActionMessage *action;
   struct acUserMovement usermove;
   LONG sticky, xchange, ychange, zchange, index;
   APTR queue;

   if (!Args) return(ObjectError(Self, ERH_UserMovement, ERR_Args));

   /* Scan ahead to see if there are more incoming UserMovement messages.
   ** If so, cache the current movement request and make use of it later.
   */

   if (AccessMemory(GetResource(RES_MESSAGEQUEUE), MEM_READ, &queue) IS ERR_Okay) {
      index = 0;
      while (ScanMessages(queue, &index, MSGID_ACTION, buffer, sizeof(buffer)) IS ERR_Okay) {
         action = (struct ActionMessage *)(buffer + sizeof(struct Message));
         if ((action->ActionID IS AC_UserClick) AND (action->ObjectID IS Self->Head.UniqueID)) {
            break;
         }
         else if ((action->ActionID IS AC_UserClickRelease) AND (action->ObjectID IS Self->Head.UniqueID)) {
            break;
         }
         else if ((action->ActionID IS AC_UserMovement) AND (action->ObjectID IS Self->Head.UniqueID)) {
            Self->prvAddX += Args->XChange;
            Self->prvAddY += Args->YChange;
            Self->prvAddZ += Args->ZChange;
            ReleaseMemory(queue);
            return(ERR_Okay|ERF_Notified);
         }
      }
      ReleaseMemory(queue);
   }
   else DPrintF("UserMovement:","[Render:%d] Failed to access my process' message queue.");

   xchange = Args->XChange + Self->prvAddX;
   ychange = Args->YChange + Self->prvAddY;
   zchange = Args->ZChange + Self->prvAddZ;

   /*** Dragging support ***/

   if (Self->prvAnchored) {
      sticky       = Self->Flags & RNF_STICKY;
      Self->Flags &= ~RNF_STICKY; /* Turn off the sticky flag, as it prevents movement */

      acMove(Self, xchange, ychange, zchange);

      if (sticky) {
         Self->Flags |= RNF_STICKY;
         UpdateRenderList(Self); /* (Required to put back the sticky flag) */
      }
   }

   /* Make a note as to wether or not the pointer is within the render object */

   if (Args->OverObject IS Self->Head.UniqueID) Self->prvPointerWithin = TRUE;
   else Self->prvPointerWithin = FALSE;

   Self->prvAddX = 0;
   Self->prvAddY = 0;
   Self->prvAddZ = 0;

   usermove.OverObject = Args->OverObject;
   usermove.XCoord  = Args->XCoord;
   usermove.YCoord  = Args->YCoord;
   usermove.ZCoord  = Args->ZCoord;
   usermove.XChange = xchange;
   usermove.YChange = ychange;
   usermove.ZChange = zchange;
   NotifySubscribers(Self, AC_UserMovement, &usermove, NSF_FORCEDELAY); //NULL);

   return(ERR_Okay|ERF_Notified);
}

/****************************************************************************/

#include "render_drawing.c"
#include "render_expose.c"
#include "render_fields.c"
#include "render_dimensions.c"
#include "render_move.c"
#include "render_movecontent.c"
#include "render_resize.c"
